"""Provides the EvolutionStrategyEmitter."""
import numpy as np

from ribs._utils import check_1d_shape, validate_batch_args
from ribs.emitters._emitter_base import EmitterBase
from ribs.emitters.opt import _get_es
from ribs.emitters.rankers import _get_ranker


class EvolutionStrategyEmitter(EmitterBase):
    """Adapts a distribution of solutions with an ES.

    This emitter originates in `Fontaine 2020
    <https://arxiv.org/abs/1912.02400>`_. The multivariate Gaussian solution
    distribution begins at ``x0`` with standard deviation ``sigma0``. Based on
    how the generated solutions are ranked (see ``ranker``), the ES then adapts
    the mean and covariance of the distribution.

    Args:
        archive (ribs.archives.ArchiveBase): An archive to use when creating
            and inserting solutions. For instance, this can be
            :class:`ribs.archives.GridArchive`.
        x0 (np.ndarray): Initial solution. Must be 1-dimensional.
        sigma0 (float): Initial step size / standard deviation.
        ranker (Callable or str): The ranker is a :class:`RankerBase` object
            that orders the solutions after they have been evaluated in the
            environment. This parameter may be a callable (e.g. a class or
            a lambda function) that takes in no parameters and returns an
            instance of :class:`RankerBase`, or it may be a full or abbreviated
            ranker name as described in
            :meth:`ribs.emitters.rankers.get_ranker`.
        es (Callable or str): The evolution strategy is an
            :class:`EvolutionStrategyBase` object that is used to adapt the
            distribution from which new solutions are sampled. This parameter
            may be a callable (e.g. a class or a lambda function) that takes in
            the parameters of :class:`EvolutionStrategyBase` along with kwargs
            provided by the ``es_kwargs`` argument, or it may be a full or
            abbreviated optimizer name as described in :mod:`ribs.emitters.opt`.
        es_kwargs (dict): Additional arguments to pass to the evolution
            strategy optimizer. See the evolution-strategy-based optimizers in
            :mod:`ribs.emitters.opt` for the arguments allowed by each
            optimizer.
        selection_rule ("mu" or "filter"): Method for selecting parents for the
            evolution strategy. With "mu" selection, the first half of the
            solutions will be selected as parents, while in "filter", any
            solutions that were added to the archive will be selected.
        restart_rule (int, "no_improvement", and "basic"): Method to use when
            checking for restarts. If given an integer, then the emitter will
            restart after this many iterations, where each iteration is a call
            to :meth:`tell`. With "basic", only the default CMA-ES convergence
            rules will be used, while with "no_improvement", the emitter will
            restart when none of the proposed solutions were added to the
            archive.
        bounds (None or array-like): Bounds of the solution space. As suggested
            in `Biedrzycki 2020
            <https://www.sciencedirect.com/science/article/abs/pii/S2210650219301622>`_,
            solutions are resampled until they fall within these bounds.  Pass
            None to indicate there are no bounds. Alternatively, pass an
            array-like to specify the bounds for each dim. Each element in this
            array-like can be None to indicate no bound, or a tuple of
            ``(lower_bound, upper_bound)``, where ``lower_bound`` or
            ``upper_bound`` may be None to indicate no bound.
        batch_size (int): Number of solutions to return in :meth:`ask`. If not
            passed in, a batch size will be automatically calculated using the
            default CMA-ES rules.
        seed (int): Value to seed the random number generator. Set to None to
            avoid a fixed seed.
    Raises:
        ValueError: There is an error in x0 or initial_solutions.
        ValueError: There is an error in the bounds configuration.
        ValueError: If ``restart_rule``, ``selection_rule``, or ``ranker`` is
            invalid.
    """

    def __init__(
        self,
        archive,
        *,
        x0,
        sigma0,
        ranker="2imp",
        es="cma_es",
        es_kwargs=None,
        selection_rule="filter",
        restart_rule="no_improvement",
        bounds=None,
        batch_size=None,
        seed=None,
    ):
        self._rng = np.random.default_rng(seed)
        self._x0 = np.array(x0, dtype=archive.dtype)
        check_1d_shape(self._x0, "x0", archive.solution_dim,
                       "archive.solution_dim")
        self._sigma0 = sigma0
        EmitterBase.__init__(
            self,
            archive,
            solution_dim=archive.solution_dim,
            bounds=bounds,
        )

        if selection_rule not in ["mu", "filter"]:
            raise ValueError(f"Invalid selection_rule {selection_rule}")
        self._selection_rule = selection_rule

        self._restart_rule = restart_rule
        self._restarts = 0
        self._itrs = 0

        # Check if the restart_rule is valid, discard check_restart result.
        _ = self._check_restart(0)

        opt_seed = None if seed is None else self._rng.integers(10_000)
        self._opt = _get_es(es,
                            sigma0=sigma0,
                            batch_size=batch_size,
                            solution_dim=self._solution_dim,
                            seed=opt_seed,
                            dtype=self.archive.dtype,
                            **(es_kwargs if es_kwargs is not None else {}))
        self._opt.reset(self._x0)

        self._ranker = _get_ranker(ranker)
        self._ranker.reset(self, archive, self._rng)

        self._batch_size = self._opt.batch_size

    @property
    def x0(self):
        """numpy.ndarray: Initial solution for the optimizer."""
        return self._x0

    @property
    def batch_size(self):
        """int: Number of solutions to return in :meth:`ask`."""
        return self._batch_size

    @property
    def restarts(self):
        """int: The number of restarts for this emitter."""
        return self._restarts

    @property
    def itrs(self):
        """int: The number of iterations for this emitter, where each iteration
        is a call to :meth:`tell`."""
        return self._itrs

    def ask(self):
        """Samples new solutions from a multivariate Gaussian.

        The multivariate Gaussian is parameterized by the evolution strategy
        optimizer ``self._opt``.

        Returns:
            (batch_size, :attr:`solution_dim`) array -- a batch of new solutions
            to evaluate.
        """
        return self._opt.ask(self.lower_bounds, self.upper_bounds)

    def _check_restart(self, num_parents):
        """Emitter-side checks for restarting the optimizer.

        The optimizer also has its own checks.

        Args:
            num_parents (int): The number of solution to propagate to the next
                generation from the solutions generated by CMA-ES.

        Raises:
          ValueError: If :attr:`restart_rule` is invalid.
        """
        if isinstance(self._restart_rule, (int, np.integer)):
            return self._itrs % self._restart_rule == 0
        if self._restart_rule == "no_improvement":
            return num_parents == 0
        if self._restart_rule == "basic":
            return False
        raise ValueError(f"Invalid restart_rule {self._restart_rule}")

    def tell(self,
             solution_batch,
             objective_batch,
             measures_batch,
             status_batch,
             value_batch,
             metadata_batch=None):
        """Gives the emitter results from evaluating solutions.

        The solutions are ranked based on the `rank()` function defined by
        `self._ranker`. Then, the ranked solutions are passed to CMA-ES for
        adaptation.

        This function also checks for restart condition and restarts CMA-ES
        when needed.

        Args:
            solution_batch (array-like): (batch_size, :attr:`solution_dim`)
                array of solutions generated by this emitter's :meth:`ask()`
                method.
            objective_batch (array-like): 1D array containing the objective
                function value of each solution.
            measures_batch (array-like): (batch_size, measure space
                dimension) array with the measure space coordinates of each
                solution.
            status_batch (array-like): 1D array of
                :class:`ribs.archive.AddStatus` returned by a series of calls
                to archive's :meth:`add()` method.
            value_batch (array-like): 1D array of floats returned by a series
                of calls to archive's :meth:`add()` method. For what these
                floats represent, refer to :meth:`ribs.archives.add()`.
            metadata_batch (array-like): 1D object array containing a metadata
                object for each solution.
        """
        # Preprocessing arguments.
        solution_batch = np.asarray(solution_batch)
        objective_batch = np.asarray(objective_batch)
        measures_batch = np.asarray(measures_batch)
        status_batch = np.asarray(status_batch)
        value_batch = np.asarray(value_batch)
        batch_size = solution_batch.shape[0]
        metadata_batch = (np.empty(batch_size, dtype=object) if
                          metadata_batch is None else np.asarray(metadata_batch,
                                                                 dtype=object))

        # Validate arguments.
        validate_batch_args(archive=self.archive,
                            solution_batch=solution_batch,
                            objective_batch=objective_batch,
                            measures_batch=measures_batch,
                            status_batch=status_batch,
                            value_batch=value_batch,
                            metadata_batch=metadata_batch)

        # Increase iteration counter.
        self._itrs += 1

        # Count number of new solutions.
        new_sols = status_batch.astype(bool).sum()

        # Sort the solutions using ranker.
        indices, ranking_values = self._ranker.rank(
            self, self.archive, self._rng, solution_batch, objective_batch,
            measures_batch, status_batch, value_batch, metadata_batch)

        # Select the number of parents.
        num_parents = (new_sols if self._selection_rule == "filter" else
                       self._batch_size // 2)

        # Update Evolution Strategy.
        self._opt.tell(indices, num_parents)

        # Check for reset.
        if (self._opt.check_stop(ranking_values[indices]) or
                self._check_restart(new_sols)):
            new_x0 = self.archive.sample_elites(1).solution_batch[0]
            self._opt.reset(new_x0)
            self._ranker.reset(self, self.archive, self._rng)
            self._restarts += 1



"""Provides ArchiveBase."""
from abc import ABC, abstractmethod
from collections import OrderedDict

import numpy as np
from numpy_groupies import aggregate_nb as aggregate

from ribs._utils import (check_1d_shape, check_batch_shape, check_finite,
                         check_is_1d, readonly, validate_batch_args,
                         validate_single_args)
from ribs.archives._archive_data_frame import ArchiveDataFrame
from ribs.archives._archive_stats import ArchiveStats
from ribs.archives._cqd_score_result import CQDScoreResult
from ribs.archives._elite import Elite, EliteBatch

_ADD_WARNING = (" Note that starting in pyribs 0.5.0, add() takes in a "
                "batch of solutions unlike in pyribs 0.4.0, where add() "
                "only took in a single solution.")


class ArchiveIterator:
    """An iterator for an archive's elites."""

    # pylint: disable = protected-access

    def __init__(self, archive):
        self.archive = archive
        self.iter_idx = 0
        self.state = archive._state.copy()

    def __iter__(self):
        """This is the iterator, so it returns itself."""
        return self

    def __next__(self):
        """Raises RuntimeError if the archive was modified with add() or
        clear()."""
        if self.state != self.archive._state:
            # This check should go first because a call to clear() would clear
            # _occupied_indices and cause StopIteration to happen early.
            raise RuntimeError(
                "Archive was modified with add() or clear() during iteration.")
        if self.iter_idx >= len(self.archive):
            raise StopIteration

        idx = self.archive._occupied_indices[self.iter_idx]
        self.iter_idx += 1
        return Elite(
            self.archive._solution_arr[idx],
            self.archive._objective_arr[idx],
            self.archive._measures_arr[idx],
            idx,
            self.archive._metadata_arr[idx],
        )


class ArchiveBase(ABC):  # pylint: disable = too-many-instance-attributes
    """Base class for archives.

    This class assumes all archives use a fixed-size container with cells that
    hold (1) information about whether the cell is occupied (bool), (2) a
    solution (1D array), (3) objective function evaluation of the solution
    (float), (4) measure space coordinates of the solution (1D array), (5)
    any additional metadata associated with the solution (object), and (6) a
    threshold which determines how high an objective value must be for a
    solution to be inserted into a cell (float). In this class, the container is
    implemented with separate numpy arrays that share common dimensions. Using
    the ``solution_dim``, ``cells`, and ``measure_dim`` arguments in
    ``__init__``, these arrays are as follows:

    +------------------------+----------------------------+
    | Name                   |  Shape                     |
    +========================+============================+
    | ``_occupied_arr``      |  ``(cells,)``              |
    +------------------------+----------------------------+
    | ``_solution_arr``      |  ``(cells, solution_dim)`` |
    +------------------------+----------------------------+
    | ``_objective_arr``     |  ``(cells,)``              |
    +------------------------+----------------------------+
    | ``_measures_arr``      |  ``(cells, measure_dim)``  |
    +------------------------+----------------------------+
    | ``_metadata_arr``      |  ``(cells,)``              |
    +------------------------+----------------------------+
    | ``_threshold_arr``     |  ``(cells,)``              |
    +------------------------+----------------------------+

    All of these arrays are accessed via a common integer index. If we have
    index ``i``, we access its solution at ``_solution_arr[i]``, its measure
    values at ``_measures_arr[i]``, etc.

    Thus, child classes typically override the following methods:

    - ``__init__``: Child classes must invoke this class's ``__init__`` with the
      appropriate arguments.
    - :meth:`index_of`: Returns integer indices into the arrays above when
      given a batch of measures. Usually, each index has a meaning, e.g. in
      :class:`~ribs.archives.CVTArchive` it is the index of a centroid.
      Documentation for this method should describe the meaning of the index.

    .. note:: Attributes beginning with an underscore are only intended to be
        accessed by child classes (i.e. they are "protected" attributes).

    .. note:: The idea of archive thresholds was introduced in `Fontaine 2022
        <https://arxiv.org/abs/2205.10752>`_. Refer to our `CMA-MAE tutorial
        <../../tutorials/cma_mae.html>`_ for more info on thresholds, including
        the ``learning_rate`` and ``threshold_min`` parameters.

    Args:
        solution_dim (int): Dimension of the solution space.
        cells (int): Number of cells in the archive. This is used to create the
            numpy arrays described above for storing archive info.
        measure_dim (int): The dimension of the measure space.
        learning_rate (float): The learning rate for threshold updates.
        threshold_min (float): The initial threshold value for all the cells.
        qd_score_offset (float): Archives often contain negative objective
            values, and if the QD score were to be computed with these negative
            objectives, the algorithm would be penalized for adding new cells
            with negative objectives. Thus, a standard practice is to normalize
            all the objectives so that they are non-negative by introducing an
            offset. This QD score offset will be *subtracted* from all
            objectives in the archive, e.g., if your objectives go as low as
            -300, pass in -300 so that each objective will be transformed as
            ``objective - (-300)``.
        seed (int): Value to seed the random number generator. Set to None to
            avoid a fixed seed.
        dtype (str or data-type): Data type of the solutions, objectives,
            and measures. We only support ``"f"`` / ``np.float32`` and ``"d"`` /
            ``np.float64``.
    Attributes:
        _solution_dim (int): See ``solution_dim`` arg.
        _rng (numpy.random.Generator): Random number generator, used in
            particular for generating random elites.
        _cells (int): See ``cells`` arg.
        _measure_dim (int): See ``measure_dim`` arg.
        _occupied_arr (numpy.ndarray): Bool array storing whether each cell in
            the archive is occupied.
        _solution_arr (numpy.ndarray): Float array storing the solutions
            themselves.
        _objective_arr (numpy.ndarray): Float array storing the objective value
            of each solution.
        _measures_arr (numpy.ndarray): Float array storing the measure space
            coordinates of each solution.
        _metadata_arr (numpy.ndarray): Object array storing the metadata
            associated with each solution.
        _threshold_arr (numpy.ndarray): Float array storing the threshold for
            insertion into each cell.
        _occupied_indices (numpy.ndarray): A ``(cells,)`` array of integer
            (``np.int32``) indices that are occupied in the archive. This could
            be a list, but for efficiency, we make it a fixed-size array, where
            only the first ``_num_occupied`` entries are valid.
        _num_occupied (int): Number of elites currently in the archive. This is
            used to index into ``_occupied_indices``.
    """

    def __init__(self,
                 *,
                 solution_dim,
                 cells,
                 measure_dim,
                 learning_rate=1.0,
                 threshold_min=-np.inf,
                 qd_score_offset=0.0,
                 seed=None,
                 dtype=np.float64):

        ## Intended to be accessed by child classes. ##
        self._solution_dim = solution_dim
        self._rng = np.random.default_rng(seed)
        self._cells = cells
        self._measure_dim = measure_dim
        self._dtype = self._parse_dtype(dtype)

        self._num_occupied = 0
        self._occupied_arr = np.zeros(self._cells, dtype=bool)
        self._occupied_indices = np.empty(self._cells, dtype=np.int32)

        self._solution_arr = np.empty((self._cells, solution_dim),
                                      dtype=self.dtype)
        self._objective_arr = np.empty(self._cells, dtype=self.dtype)
        self._measures_arr = np.empty((self._cells, self._measure_dim),
                                      dtype=self.dtype)
        self._metadata_arr = np.empty(self._cells, dtype=object)

        if threshold_min == -np.inf and learning_rate != 1.0:
            raise ValueError("threshold_min can only be -np.inf if "
                             "learning_rate is 1.0")
        self._learning_rate = self._dtype(learning_rate)
        self._threshold_min = self._dtype(threshold_min)
        self._threshold_arr = np.full(self._cells,
                                      threshold_min,
                                      dtype=self.dtype)

        self._qd_score_offset = self._dtype(qd_score_offset)

        self._stats = None
        # Sum of all objective values in the archive; useful for computing
        # qd_score and obj_mean.
        self._objective_sum = None
        self._stats_reset()

        self._best_elite = None

        # Tracks archive modifications by counting calls to clear() and add().
        self._state = {"clear": 0, "add": 0}

        ## Not intended to be accessed by children. ##
        self._seed = seed

    @staticmethod
    def _parse_dtype(dtype):
        """Parses the dtype passed into the constructor.

        Returns:
            np.float32 or np.float64
        Raises:
            ValueError: There is an error in the bounds configuration.
        """
        # First convert str dtype's to np.dtype.
        if isinstance(dtype, str):
            dtype = np.dtype(dtype)

        # np.dtype is not np.float32 or np.float64, but it compares equal.
        if dtype == np.float32:
            return np.float32
        if dtype == np.float64:
            return np.float64

        raise ValueError("Unsupported dtype. Must be np.float32 or np.float64")

    @property
    def cells(self):
        """int: Total number of cells in the archive."""
        return self._cells

    @property
    def empty(self):
        """bool: Whether the archive is empty."""
        return self._num_occupied == 0

    @property
    def measure_dim(self):
        """int: Dimensionality of the measure space."""
        return self._measure_dim

    @property
    def solution_dim(self):
        """int: Dimensionality of the solutions in the archive."""
        return self._solution_dim

    @property
    def learning_rate(self):
        """float: The learning rate for threshold updates."""
        return self._learning_rate

    @property
    def threshold_min(self):
        """float: The initial threshold value for all the cells."""
        return self._threshold_min

    @property
    def qd_score_offset(self):
        """float: The offset which is subtracted from objective values when
        computing the QD score."""
        return self._qd_score_offset

    @property
    def stats(self):
        """:class:`ArchiveStats`: Statistics about the archive.

        See :class:`ArchiveStats` for more info.
        """
        return self._stats

    @property
    def best_elite(self):
        """:class:`Elite`: The elite with the highest objective in the archive.

        None if there are no elites in the archive.

        .. note::
            If the archive is non-elitist (this occurs when using the archive
            with a learning rate which is not 1.0, as in CMA-MAE), then this
            best elite may no longer exist in the archive because it was
            replaced with an elite with a lower objective value. This can happen
            because in non-elitist archives, new solutions only need to exceed
            the *threshold* of the cell they are being inserted into, not the
            *objective* of the elite currently in the cell. See `#314
            <https://github.com/icaros-usc/pyribs/pull/314>`_ for more info.
        """
        return self._best_elite

    @property
    def dtype(self):
        """data-type: The dtype of the solutions, objective, and measures."""
        return self._dtype

    def __len__(self):
        """Number of elites in the archive."""
        return self._num_occupied

    def __iter__(self):
        """Creates an iterator over the :class:`Elite`'s in the archive.

        Example:

            ::

                for elite in archive:
                    elite.sol
                    elite.obj
                    ...
        """
        return ArchiveIterator(self)

    def _stats_reset(self):
        """Resets the archive stats."""
        self._stats = ArchiveStats(
            num_elites=0,
            coverage=self.dtype(0.0),
            qd_score=self.dtype(0.0),
            norm_qd_score=self.dtype(0.0),
            obj_max=None,
            obj_mean=None,
        )
        self._objective_sum = self.dtype(0.0)

    def _compute_new_thresholds(self, threshold_arr, objective_batch,
                                index_batch, learning_rate):
        """Update thresholds.

        Args:
            threshold_arr (np.ndarray): The threshold of the cells before
                updating. 1D array.
            objective_batch (np.ndarray): The objective values of the solution
                that is inserted into the archive for each cell. 1D array. We
                assume that the objective values are all higher than the
                thresholds of their respective cells.
            index_batch (np.ndarray): The archive index of the elements in
                objective batch.
        Returns:
            `new_threshold_batch` (A self.dtype array of new
            thresholds) and `threshold_update_indices` (A boolean
            array indicating which entries in `threshold_arr` should
            be updated.
        """
        # Even though we do this check, it should not be possible to have
        # empty objective_batch or index_batch in the add() method since
        # we check that at least one cell is being updated by seeing if
        # can_insert has any True values.
        if objective_batch.size == 0 or index_batch.size == 0:
            return np.array([], dtype=self.dtype), np.array([], dtype=bool)

        # Compute the number of objectives inserted into each cell.
        objective_sizes = aggregate(index_batch,
                                    objective_batch,
                                    func="len",
                                    fill_value=0,
                                    size=threshold_arr.size)

        # These indices are with respect to the archive, so we can directly pass
        # them to threshold_arr.
        threshold_update_indices = objective_sizes > 0

        # Compute the sum of the objectives inserted into each cell.
        objective_sums = aggregate(index_batch,
                                   objective_batch,
                                   func="sum",
                                   fill_value=np.nan,
                                   size=threshold_arr.size)

        # Throw away indices that we do not care about.
        objective_sizes = objective_sizes[threshold_update_indices]
        objective_sums = objective_sums[threshold_update_indices]

        # Unlike in add_single, we do not need to worry about
        # old_threshold having -np.inf here as a result of threshold_min
        # being -np.inf. This is because the case with threshold_min =
        # -np.inf is handled separately since we compute the new
        # threshold based on the max objective in each cell in that case.
        old_threshold = np.copy(threshold_arr[threshold_update_indices])

        ratio = self.dtype(1.0 - learning_rate)**objective_sizes
        new_threshold_batch = (ratio * old_threshold +
                               (objective_sums / objective_sizes) * (1 - ratio))

        return new_threshold_batch, threshold_update_indices

    def clear(self):
        """Removes all elites from the archive.

        After this method is called, the archive will be :attr:`empty`.
        """
        # Clear ``self._occupied_indices`` and ``self._occupied_arr`` since a
        # cell can have arbitrary values when its index is marked as unoccupied.
        self._num_occupied = 0  # Corresponds to clearing _occupied_indices.
        self._occupied_arr.fill(False)

        # We also need to reset thresholds since archive addition is based on
        # thresholds.
        self._threshold_arr.fill(self._threshold_min)

        self._state["clear"] += 1
        self._state["add"] = 0

        self._stats_reset()
        self._best_elite = None

    @abstractmethod
    def index_of(self, measures_batch):
        """Returns archive indices for the given batch of measures.

        If you need to retrieve the index of the measures for a *single*
        solution, consider using :meth:`index_of_single`.

        Args:
            measures_batch (array-like): (batch_size, :attr:`measure_dim`)
                array of coordinates in measure space.
        Returns:
            (numpy.ndarray): (batch_size,) array with the indices of the
            batch of measures in the archive's storage arrays.
        """

    def index_of_single(self, measures):
        """Returns the index of the measures for one solution.

        While :meth:`index_of` takes in a *batch* of measures, this method takes
        in the measures for only *one* solution. If :meth:`index_of` is
        implemented correctly, this method should work immediately (i.e. `"out
        of the box" <https://idioms.thefreedictionary.com/Out-of-the-Box>`_).

        Args:
            measures (array-like): (:attr:`measure_dim`,) array of measures for
                a single solution.
        Returns:
            int or numpy.integer: Integer index of the measures in the archive's
            storage arrays.
        Raises:
            ValueError: ``measures`` is not of shape (:attr:`measure_dim`,).
            ValueError: ``measures`` has non-finite values (inf or NaN).
        """
        measures = np.asarray(measures)
        check_1d_shape(measures, "measures", self.measure_dim, "measure_dim")
        check_finite(measures, "measures")
        return self.index_of(measures[None])[0]

    def add(self,
            solution_batch,
            objective_batch,
            measures_batch,
            metadata_batch=None):
        """Inserts a batch of solutions into the archive.

        Each solution is only inserted if it has a higher ``objective`` than the
        threshold of the corresponding cell. For the default values of
        ``learning_rate`` and ``threshold_min``, this threshold is simply the
        objective value of the elite previously in the cell.  If multiple
        solutions in the batch end up in the same cell, we only insert the
        solution with the highest objective. If multiple solutions end up in the
        same cell and tie for the highest objective, we insert the solution that
        appears first in the batch.

        For the default values of ``learning_rate`` and ``threshold_min``, the
        threshold for each cell is updated by taking the maximum objective value
        among all the solutions that landed in the cell, resulting in the same
        behavior as in the vanilla MAP-Elites archive. However, for other
        settings, the threshold is updated with the batch update rule described
        in the appendix of `Fontaine 2022 <https://arxiv.org/abs/2205.10752>`_.

        .. note:: The indices of all arguments should "correspond" to each
            other, i.e. ``solution_batch[i]``, ``objective_batch[i]``,
            ``measures_batch[i]``, and ``metadata_batch[i]`` should be the
            solution parameters, objective, measures, and metadata for solution
            ``i``.

        Args:
            solution_batch (array-like): (batch_size, :attr:`solution_dim`)
                array of solution parameters.
            objective_batch (array-like): (batch_size,) array with objective
                function evaluations of the solutions.
            measures_batch (array-like): (batch_size, :attr:`measure_dim`)
                array with measure space coordinates of all the solutions.
            metadata_batch (array-like): (batch_size,) array of Python objects
                representing metadata for the solution. For instance, this could
                be a dict with several properties.

                .. warning:: Due to how NumPy's :func:`~numpy.asarray`
                    automatically converts array-like objects to arrays, passing
                    array-like objects as metadata may lead to unexpected
                    behavior. However, the metadata may be a dict or other
                    object which *contains* arrays, i.e. ``metadata_batch``
                    could be an array of dicts which contain arrays.
        Returns:
            tuple: 2-element tuple of (status_batch, value_batch) which
            describes the results of the additions. These outputs are
            particularly useful for algorithms such as CMA-ME.

            - **status_batch** (:class:`numpy.ndarray` of :class:`int`): An
              array of integers which represent the "status" obtained when
              attempting to insert each solution in the batch. Each item has the
              following possible values:

              - ``0``: The solution was not added to the archive.
              - ``1``: The solution improved the objective value of a cell
                which was already in the archive.
              - ``2``: The solution discovered a new cell in the archive.

              All statuses (and values, below) are computed with respect to the
              *current* archive. For example, if two solutions both introduce
              the same new archive cell, then both will be marked with ``2``.

              The alternative is to depend on the order of the solutions in the
              batch -- for example, if we have two solutions ``a`` and ``b``
              which introduce the same new cell in the archive, ``a`` could be
              inserted first with status ``2``, and ``b`` could be inserted
              second with status ``1`` because it improves upon ``a``. However,
              our implementation does **not** do this.

              To convert statuses to a more semantic format, cast all statuses
              to :class:`AddStatus` e.g. with ``[AddStatus(s) for s in
              status_batch]``.

            - **value_batch** (:attr:`dtype`): An array with values for each
              solution in the batch. With the default values of ``learning_rate
              = 1.0`` and ``threshold_min = -np.inf``, the meaning of each value
              depends on the corresponding ``status`` and is identical to that
              in CMA-ME (`Fontaine 2020 <https://arxiv.org/abs/1912.02400>`_):

              - ``0`` (not added): The value is the "negative improvement," i.e.
                the objective of the solution passed in minus the objective of
                the elite still in the archive (this value is negative because
                the solution did not have a high enough objective to be added to
                the archive).
              - ``1`` (improve existing cell): The value is the "improvement,"
                i.e. the objective of the solution passed in minus the objective
                of the elite previously in the archive.
              - ``2`` (new cell): The value is just the objective of the
                solution.

              In contrast, for other values of ``learning_rate`` and
              ``threshold_min``, each value is equivalent to the objective value
              of the solution minus the threshold of its corresponding cell in
              the archive.
        Raises:
            ValueError: The array arguments do not match their specified shapes.
            ValueError: ``objective_batch`` or ``measures_batch`` has non-finite
                values (inf or NaN).
        """
        self._state["add"] += 1

        ## Step 0: Preprocess input. ##
        solution_batch = np.asarray(solution_batch)
        objective_batch = np.asarray(objective_batch)
        measures_batch = np.asarray(measures_batch)
        batch_size = solution_batch.shape[0]
        metadata_batch = (np.empty(batch_size, dtype=object) if
                          metadata_batch is None else np.asarray(metadata_batch,
                                                                 dtype=object))

        ## Step 1: Validate input. ##
        validate_batch_args(
            archive=self,
            solution_batch=solution_batch,
            objective_batch=objective_batch,
            measures_batch=measures_batch,
            metadata_batch=metadata_batch,
        )

        ## Step 2: Compute status_batch and value_batch ##

        # Retrieve indices.
        index_batch = self.index_of(measures_batch)

        # Copy old objectives since we will be modifying the objectives storage.
        old_objective_batch = np.copy(self._objective_arr[index_batch])
        old_threshold_batch = np.copy(self._threshold_arr[index_batch])

        # Compute the statuses -- these are all boolean arrays of length
        # batch_size.
        already_occupied = self._occupied_arr[index_batch]
        # In the case where we want CMA-ME behavior, threshold_arr[index]
        # is -inf for new cells, which satisfies the condition for can_be_added.
        can_be_added = objective_batch > old_threshold_batch
        is_new = can_be_added & ~already_occupied
        improve_existing = can_be_added & already_occupied
        status_batch = np.zeros(batch_size, dtype=np.int32)
        status_batch[is_new] = 2
        status_batch[improve_existing] = 1

        # New solutions require special settings for old_objective and
        # old_threshold.
        old_objective_batch[is_new] = self.dtype(0)

        # If threshold_min is -inf, then we want CMA-ME behavior, which
        # will compute the improvement value of new solutions w.r.t zero.
        # Otherwise, we will compute w.r.t. threshold_min.
        old_threshold_batch[is_new] = (self.dtype(0) if self._threshold_min
                                       == -np.inf else self._threshold_min)
        value_batch = objective_batch - old_threshold_batch

        ## Step 3: Insert solutions into archive. ##

        # Return early if we cannot insert anything -- continuing would actually
        # throw a ValueError in aggregate() since index_batch[can_insert] would
        # be empty.
        can_insert = is_new | improve_existing
        if not np.any(can_insert):
            return status_batch, value_batch

        # Select only solutions that can be inserted into the archive.
        solution_batch_can = solution_batch[can_insert]
        objective_batch_can = objective_batch[can_insert]
        measures_batch_can = measures_batch[can_insert]
        index_batch_can = index_batch[can_insert]
        metadata_batch_can = metadata_batch[can_insert]
        old_objective_batch_can = old_objective_batch[can_insert]

        # Retrieve indices of solutions that should be inserted into the
        # archive. Currently, multiple solutions may be inserted at each
        # archive index, but we only want to insert the maximum among these
        # solutions. Thus, we obtain the argmax for each archive index.
        #
        # We use a fill_value of -1 to indicate archive indices which were not
        # covered in the batch. Note that the length of archive_argmax is only
        # max(index_batch[can_insert]), rather than the total number of grid
        # cells. However, this is okay because we only need the indices of the
        # solutions, which we store in should_insert.
        #
        # aggregate() always chooses the first item if there are ties, so the
        # first elite will be inserted if there is a tie. See their default
        # numpy implementation for more info:
        # https://github.com/ml31415/numpy-groupies/blob/master/numpy_groupies/aggregate_numpy.py#L107
        archive_argmax = aggregate(index_batch_can,
                                   objective_batch_can,
                                   func="argmax",
                                   fill_value=-1)
        should_insert = archive_argmax[archive_argmax != -1]

        # Select only solutions that will be inserted into the archive.
        solution_batch_insert = solution_batch_can[should_insert]
        objective_batch_insert = objective_batch_can[should_insert]
        measures_batch_insert = measures_batch_can[should_insert]
        index_batch_insert = index_batch_can[should_insert]
        metadata_batch_insert = metadata_batch_can[should_insert]
        old_objective_batch_insert = old_objective_batch_can[should_insert]

        # Set archive storage.
        self._objective_arr[index_batch_insert] = objective_batch_insert
        self._measures_arr[index_batch_insert] = measures_batch_insert
        self._solution_arr[index_batch_insert] = solution_batch_insert
        self._metadata_arr[index_batch_insert] = metadata_batch_insert
        self._occupied_arr[index_batch_insert] = True

        # Mark new indices as occupied.
        is_new_and_inserted = is_new[can_insert][should_insert]
        n_new = np.sum(is_new_and_inserted)
        self._occupied_indices[self._num_occupied:self._num_occupied +
                               n_new] = (
                                   index_batch_insert[is_new_and_inserted])
        self._num_occupied += n_new

        # Update the thresholds.
        if self._threshold_min == -np.inf:
            # Here we want regular archive behavior, so the thresholds
            # should just be the maximum objective.
            self._threshold_arr[index_batch_insert] = objective_batch_insert
        else:
            # Here we compute the batch threshold update described in the
            # appendix of Fontaine 2022 https://arxiv.org/abs/2205.10752
            # This computation is based on the mean objective of all
            # solutions in the batch that could have been inserted into
            # each cell. This method is separated out to facilitate
            # testing.
            (new_thresholds,
             update_thresholds_indices) = self._compute_new_thresholds(
                 self._threshold_arr, objective_batch_can, index_batch_can,
                 self._learning_rate)
            self._threshold_arr[update_thresholds_indices] = new_thresholds

        ## Step 4: Update archive stats. ##

        # Since we set the new solutions in the old objective batch to have
        # value 0.0, the objectives for new solutions are added in properly
        # here.
        self._objective_sum += np.sum(objective_batch_insert -
                                      old_objective_batch_insert)
        new_qd_score = (self._objective_sum -
                        self.dtype(len(self)) * self._qd_score_offset)
        max_idx = np.argmax(objective_batch_insert)
        max_obj_insert = objective_batch_insert[max_idx]

        if self._stats.obj_max is None or max_obj_insert > self._stats.obj_max:
            new_obj_max = max_obj_insert
            self._best_elite = Elite(
                readonly(np.copy(solution_batch_insert[max_idx])),
                objective_batch_insert[max_idx],
                readonly(np.copy(measures_batch_insert[max_idx])),
                index_batch_insert[max_idx],
                metadata_batch_insert[max_idx],
            )
        else:
            new_obj_max = self._stats.obj_max

        self._stats = ArchiveStats(
            num_elites=len(self),
            coverage=self.dtype(len(self) / self.cells),
            qd_score=new_qd_score,
            norm_qd_score=self.dtype(new_qd_score / self.cells),
            obj_max=new_obj_max,
            obj_mean=self._objective_sum / self.dtype(len(self)),
        )

        return status_batch, value_batch

    def add_single(self, solution, objective, measures, metadata=None):
        """Inserts a single solution into the archive.

        The solution is only inserted if it has a higher ``objective`` than the
        threshold of the corresponding cell. For the default values of
        ``learning_rate`` and ``threshold_min``, this threshold is simply the
        objective value of the elite previously in the cell.  The threshold is
        also updated if the solution was inserted.

        .. note::
            To make it more amenable to modifications, this method's
            implementation is designed to be readable at the cost of
            performance, e.g., none of its operations are modified. If you need
            performance, we recommend using :meth:`add`.

        Args:
            solution (array-like): Parameters of the solution.
            objective (float): Objective function evaluation of the solution.
            measures (array-like): Coordinates in measure space of the solution.
            metadata (object): Python object representing metadata for the
                solution. For instance, this could be a dict with several
                properties.

                .. warning:: Due to how NumPy's :func:`~numpy.asarray`
                    automatically converts array-like objects to arrays, passing
                    array-like objects as metadata may lead to unexpected
                    behavior. However, the metadata may be a dict or other
                    object which *contains* arrays.
        Raises:
            ValueError: The array arguments do not match their specified shapes.
            ValueError: ``objective`` is non-finite (inf or NaN) or ``measures``
                has non-finite values.
        Returns:
            tuple: 2-element tuple of (status, value) describing the result of
            the add operation. Refer to :meth:`add` for the meaning of the
            status and value.
        """
        self._state["add"] += 1

        solution = np.asarray(solution)
        objective = self.dtype(objective)
        measures = np.asarray(measures)
        validate_single_args(
            self,
            solution=solution,
            objective=objective,
            measures=measures,
        )

        index = self.index_of_single(measures)

        # Only used for computing QD score.
        old_objective = self._objective_arr[index]

        # Used for computing improvement value.
        old_threshold = self._threshold_arr[index]

        # New solutions require special settings for old_objective and
        # old_threshold.
        was_occupied = self._occupied_arr[index]
        if not was_occupied:
            old_objective = self.dtype(0)
            # If threshold_min is -inf, then we want CMA-ME behavior, which will
            # compute the improvement value w.r.t. zero for new solutions.
            # Otherwise, we will compute w.r.t. threshold_min.
            old_threshold = (self.dtype(0) if self._threshold_min == -np.inf
                             else self._threshold_min)

        status = 0  # NOT_ADDED
        # In the case where we want CMA-ME behavior, threshold_arr[index]
        # is -inf for new cells, which satisfies this if condition.
        if self._threshold_arr[index] < objective:
            if was_occupied:
                status = 1  # IMPROVE_EXISTING
            else:
                # Set this index to be occupied.
                self._occupied_arr[index] = True
                self._occupied_indices[self._num_occupied] = index
                self._num_occupied += 1

                status = 2  # NEW

            # This calculation works in the case where threshold_min is -inf
            # because old_threshold will be set to 0.0 instead.
            self._threshold_arr[index] = (old_threshold *
                                          (1.0 - self._learning_rate) +
                                          objective * self._learning_rate)

            # Insert into the archive.
            self._objective_arr[index] = objective
            self._measures_arr[index] = measures
            self._solution_arr[index] = solution
            self._metadata_arr[index] = metadata

        if status:
            # Update archive stats.
            self._objective_sum += objective - old_objective
            new_qd_score = (self._objective_sum -
                            self.dtype(len(self)) * self._qd_score_offset)

            if self._stats.obj_max is None or objective > self._stats.obj_max:
                new_obj_max = objective
                self._best_elite = Elite(
                    readonly(np.copy(self._solution_arr[index])),
                    objective,
                    readonly(np.copy(self._measures_arr[index])),
                    index,
                    metadata,
                )
            else:
                new_obj_max = self._stats.obj_max

            self._stats = ArchiveStats(
                num_elites=len(self),
                coverage=self.dtype(len(self) / self.cells),
                qd_score=new_qd_score,
                norm_qd_score=self.dtype(new_qd_score / self.cells),
                obj_max=new_obj_max,
                obj_mean=self._objective_sum / self.dtype(len(self)),
            )

        return status, objective - old_threshold

    def retrieve(self, measures_batch):
        """Retrieves the elites with measures in the same cells as the measures
        specified.

        This method operates in batch, i.e. it takes in a batch of measures and
        outputs an :namedtuple:`EliteBatch`. Since :namedtuple:`EliteBatch` is a
        namedtuple, it can be unpacked::

            solution_batch, objective_batch, measures_batch, \\
                index_batch, metadata_batch = archive.retrieve(...)

        Or the fields may be accessed by name::

            elite_batch = archive.retrieve(...)
            elite_batch.solution_batch
            elite_batch.objective_batch
            elite_batch.measures_batch
            elite_batch.index_batch
            elite_batch.metadata_batch

        If the cell associated with ``measures_batch[i]`` has an elite in it,
        then ``elite_batch.solution_batch[i]``,
        ``elite_batch.objective_batch[i]``, ``elite_batch.measures_batch[i]``,
        ``elite_batch.index_batch[i]``, and ``elite_batch.metadata_batch[i]``
        will be set to the properties of the elite. Note that
        ``elite_batch.measures_batch[i]`` may not be equal to
        ``measures_batch[i]`` since the measures only need to be in the same
        archive cell.

        If the cell associated with ``measures_batch[i]`` *does not* have any
        elite in it, then the corresponding outputs are set to empty values --
        namely:

        * ``elite_batch.solution_batch[i]`` will be an array of NaN
        * ``elite_batch.objective_batch[i]`` will be NaN
        * ``elite_batch.measures_batch[i]`` will be an array of NaN
        * ``elite_batch.index_batch[i]`` will be -1
        * ``elite_batch.metadata_batch[i]`` will be None

        If you need to retrieve a *single* elite associated with some measures,
        consider using :meth:`retrieve_single`.

        Args:
            measures_batch (array-like): (batch_size, :attr:`measure_dim`)
                array of coordinates in measure space.
        Returns:
            EliteBatch: See above.
        Raises:
            ValueError: ``measures_batch`` is not of shape (batch_size,
                :attr:`measure_dim`).
            ValueError: ``measures_batch`` has non-finite values (inf or NaN).
        """
        measures_batch = np.asarray(measures_batch)
        check_batch_shape(measures_batch, "measures_batch", self.measure_dim,
                          "measure_dim")
        check_finite(measures_batch, "measures_batch")

        index_batch = self.index_of(measures_batch)
        occupied_batch = self._occupied_arr[index_batch]
        expanded_occupied_batch = occupied_batch[:, None]

        return EliteBatch(
            solution_batch=readonly(
                # For each occupied_batch[i], this np.where selects
                # self._solution_arr[index_batch][i] if occupied_batch[i] is
                # True. Otherwise, it uses the alternate value (a solution
                # array consisting of np.nan).
                np.where(
                    expanded_occupied_batch,
                    self._solution_arr[index_batch],
                    np.full(self._solution_dim, np.nan),
                )),
            objective_batch=readonly(
                np.where(
                    occupied_batch,
                    self._objective_arr[index_batch],
                    # Here the alternative is just a scalar np.nan.
                    np.nan,
                )),
            measures_batch=readonly(
                np.where(
                    expanded_occupied_batch,
                    self._measures_arr[index_batch],
                    # And here it is a measures array of np.nan.
                    np.full(self._measure_dim, np.nan),
                )),
            index_batch=readonly(
                np.where(
                    occupied_batch,
                    index_batch,
                    # Indices must be integers, so np.nan would not work, hence
                    # we use -1.
                    -1,
                )),
            metadata_batch=readonly(
                np.where(
                    occupied_batch,
                    self._metadata_arr[index_batch],
                    None,
                )),
        )

    def retrieve_single(self, measures):
        """Retrieves the elite with measures in the same cell as the measures
        specified.

        While :meth:`retrieve` takes in a *batch* of measures, this method takes
        in the measures for only *one* solution and returns a single
        :namedtuple:`Elite`.

        Args:
            measures (array-like): (:attr:`measure_dim`,) array of measures.
        Returns:
            If there is an elite with measures in the same cell as the measures
            specified, then this method returns an :namedtuple:`Elite` where all
            the fields hold the info of that elite. Otherwise, this method
            returns an :namedtuple:`Elite` filled with the same "empty" values
            described in :meth:`retrieve`.
        Raises:
            ValueError: ``measures`` is not of shape (:attr:`measure_dim`,).
            ValueError: ``measures`` has non-finite values (inf or NaN).
        """
        measures = np.asarray(measures)
        check_1d_shape(measures, "measures", self.measure_dim, "measure_dim")
        check_finite(measures, "measures")

        elite_batch = self.retrieve(measures[None])
        return Elite(
            elite_batch.solution_batch[0],
            elite_batch.objective_batch[0],
            elite_batch.measures_batch[0],
            elite_batch.index_batch[0],
            elite_batch.metadata_batch[0],
        )

    def sample_elites(self, n):
        """Randomly samples elites from the archive.

        Currently, this sampling is done uniformly at random. Furthermore, each
        sample is done independently, so elites may be repeated in the sample.
        Additional sampling methods may be supported in the future.

        Since :namedtuple:`EliteBatch` is a namedtuple, the result can be
        unpacked (here we show how to ignore some of the fields)::

            solution_batch, objective_batch, measures_batch, *_ = \\
                archive.sample_elites(32)

        Or the fields may be accessed by name::

            elite = archive.sample_elites(16)
            elite.solution_batch
            elite.objective_batch
            ...

        Args:
            n (int): Number of elites to sample.
        Returns:
            EliteBatch: A batch of elites randomly selected from the archive.
        Raises:
            IndexError: The archive is empty.
        """
        if self.empty:
            raise IndexError("No elements in archive.")

        random_indices = self._rng.integers(self._num_occupied, size=n)
        selected_indices = self._occupied_indices[random_indices]

        return EliteBatch(
            readonly(self._solution_arr[selected_indices]),
            readonly(self._objective_arr[selected_indices]),
            readonly(self._measures_arr[selected_indices]),
            readonly(selected_indices),
            readonly(self._metadata_arr[selected_indices]),
        )

    def as_pandas(self, include_solutions=True, include_metadata=False):
        """Converts the archive into an :class:`ArchiveDataFrame` (a child class
        of :class:`pandas.DataFrame`).

        The implementation of this method in :class:`ArchiveBase` creates a
        dataframe consisting of:

        - 1 column of integers (``np.int32``) for the index, named ``index``.
          See :meth:`index_of` for more info.
        - :attr:`measure_dim` columns for the measures, named ``measure_0,
          measure_1, ...``
        - 1 column for the objectives, named ``objective``
        - :attr:`solution_dim` columns for the solution parameters, named
          ``solution_0, solution_1, ...``
        - 1 column for the metadata objects, named ``metadata``

        In short, the dataframe looks like this:

        +-------+------------+------+------------+-------------+-----+----------+
        | index | measure_0  | ...  | objective  | solution_0  | ... | metadata |
        +=======+============+======+============+=============+=====+==========+
        |       |            | ...  |            |             | ... |          |
        +-------+------------+------+------------+-------------+-----+----------+

        Compared to :class:`pandas.DataFrame`, the :class:`ArchiveDataFrame`
        adds methods and attributes which make it easier to manipulate archive
        data. For more information, refer to the :class:`ArchiveDataFrame`
        documentation.

        Args:
            include_solutions (bool): Whether to include solution columns.
            include_metadata (bool): Whether to include the metadata column.
                Note that methods like :meth:`~pandas.DataFrame.to_csv` may not
                properly save the dataframe since the metadata objects may not
                be representable in a CSV.
        Returns:
            ArchiveDataFrame: See above.
        """  # pylint: disable = line-too-long
        data = OrderedDict()
        indices = self._occupied_indices[:self._num_occupied]

        # Copy indices so we do not overwrite.
        data["index"] = np.copy(indices)

        measures_batch = self._measures_arr[indices]
        for i in range(self._measure_dim):
            data[f"measure_{i}"] = measures_batch[:, i]

        data["objective"] = self._objective_arr[indices]

        if include_solutions:
            solutions = self._solution_arr[indices]
            for i in range(self._solution_dim):
                data[f"solution_{i}"] = solutions[:, i]

        if include_metadata:
            data["metadata"] = self._metadata_arr[indices]

        return ArchiveDataFrame(
            data,
            copy=False,  # Fancy indexing above already results in copying.
        )

    def cqd_score(self,
                  iterations,
                  target_points,
                  penalties,
                  obj_min,
                  obj_max,
                  dist_max=None,
                  dist_ord=None):
        """Computes the CQD score of the archive.

        The Continuous Quality Diversity (CQD) score was introduced in
        `Kent 2022 <https://dl.acm.org/doi/10.1145/3520304.3534018>`_.

        .. note:: This method by default assumes that the archive has an
            ``upper_bounds`` and ``lower_bounds`` property which delineate the
            bounds of the measure space, as is the case in
            :class:`~ribs.archives.GridArchive`,
            :class:`~ribs.archives.CVTArchive`, and
            :class:`~ribs.archives.SlidingBoundariesArchive`.  If this is not
            the case, ``dist_max`` must be passed in, and ``target_points`` must
            be an array of custom points.

        Args:
            iterations (int): Number of times to compute the CQD score. We
                return the mean CQD score across these iterations.
            target_points (int or array-like): Number of target points to
                generate, or an (iterations, n, measure_dim) array which
                lists n target points to list on each iteration. When an int is
                passed, the points are sampled uniformly within the bounds of
                the measure space.
            penalties (int or array-like): Number of penalty values over which
                to compute the score (the values are distributed evenly over the
                range [0,1]). Alternatively, this may be a 1D array which
                explicitly lists the penalty values. Known as :math:`\\theta` in
                Kent 2022.
            obj_min (float): Minimum objective value, used when normalizing the
                objectives.
            obj_max (float): Maximum objective value, used when normalizing the
                objectives.
            dist_max (float): Maximum distance between points in measure space.
                Defaults to the distance between the extremes of the measure
                space bounds (the type of distance is computed with the order
                specified by ``dist_ord``). Known as :math:`\\delta_{max}` in
                Kent 2022.
            dist_ord: Order of the norm to use for calculating measure space
                distance; this is passed to :func:`numpy.linalg.norm` as the
                ``ord`` argument. See :func:`numpy.linalg.norm` for possible
                values. The default is to use Euclidean distance (L2 norm).
        Returns:
            The mean CQD score obtained with ``iterations`` rounds of
            calculations.
        Raises:
            RuntimeError: The archive does not have the bounds properties
                mentioned above, and dist_max is not specified or the target
                points are not provided.
            ValueError: target_points or penalties is an array with the wrong
                shape.
        """
        if (not (hasattr(self, "upper_bounds") and
                 hasattr(self, "lower_bounds")) and
            (dist_max is None or np.isscalar(target_points))):
            raise RuntimeError(
                "When the archive does not have lower_bounds and "
                "upper_bounds properties, dist_max must be specified, "
                "and target_points must be an array")

        if np.isscalar(target_points):
            # pylint: disable = no-member
            target_points = self._rng.uniform(
                low=self.lower_bounds,
                high=self.upper_bounds,
                size=(iterations, target_points, self.measure_dim),
            )
        else:
            # Copy since we return this.
            target_points = np.copy(target_points)
            if (target_points.ndim != 3 or
                    target_points.shape[0] != iterations or
                    target_points.shape[2] != self.measure_dim):
                raise ValueError(
                    "Expected target_points to be a 3D array with "
                    f"shape ({iterations}, n, {self.measure_dim}) "
                    "(i.e. shape (iterations, n, measure_dim)) but it had "
                    f"shape {target_points.shape}")

        if dist_max is None:
            # pylint: disable = no-member
            dist_max = np.linalg.norm(self.upper_bounds - self.lower_bounds,
                                      ord=dist_ord)

        if np.isscalar(penalties):
            penalties = np.linspace(0, 1, penalties)
        else:
            penalties = np.copy(penalties)  # Copy since we return this.
            check_is_1d(penalties, "penalties")

        index_batch = self._occupied_indices[:self._num_occupied]
        measures_batch = self._measures_arr[index_batch]
        objective_batch = self._objective_arr[index_batch]

        norm_objectives = objective_batch / (obj_max - obj_min)

        scores = np.zeros(iterations)

        for itr in range(iterations):
            # Distance calculation -- start by taking the difference between
            # each measure i and all the target points.
            distances = measures_batch[:, None] - target_points[itr]

            # (len(archive), n_target_points) array of distances.
            distances = np.linalg.norm(distances, ord=dist_ord, axis=2)

            norm_distances = distances / dist_max

            for penalty in penalties:
                # Known as omega in Kent 2022 -- a (len(archive),
                # n_target_points) array.
                values = norm_objectives[:, None] - penalty * norm_distances

                # (n_target_points,) array.
                max_values_per_target = np.max(values, axis=0)

                scores[itr] += np.sum(max_values_per_target)

        return CQDScoreResult(
            iterations=iterations,
            mean=np.mean(scores),
            scores=scores,
            target_points=target_points,
            penalties=penalties,
            obj_min=obj_min,
            obj_max=obj_max,
            dist_max=dist_max,
            dist_ord=dist_ord,
        )



"""Provides EvolutionStrategyBase."""
from abc import ABC, abstractmethod

import numpy as np


class EvolutionStrategyBase(ABC):
    """Base class for evolution strategy optimizers for use with emitters.

    The basic usage is:

    - Initialize the optimizer and reset it.
    - Repeatedly:

        - Request new solutions with ``ask()``
        - Rank the solutions in the emitter (better solutions come first) and
          pass the indices back with ``tell()``.
        - Use ``check_stop()`` to see if the optimizer has reached a stopping
          condition, and if so, call ``reset()``.

    Args:
        sigma0 (float): Initial step size.
        solution_dim (int): Size of the solution space.
        batch_size (int): Number of solutions to evaluate at a time.
        seed (int): Seed for the random number generator.
        dtype (str or data-type): Data type of solutions.
    """

    def __init__(self,
                 sigma0,
                 solution_dim,
                 batch_size=None,
                 seed=None,
                 dtype=np.float64):
        pass

    @abstractmethod
    def reset(self, x0):
        """Resets the optimizer to start at x0.

        Args:
            x0 (np.ndarray): Initial mean.
        """

    @abstractmethod
    def check_stop(self, ranking_values):
        """Checks if the optimizer should stop and be reset.

        Args:
            ranking_values (np.ndarray): Array of objective values of the
                solutions, sorted in the same order that the solutions were
                sorted when passed to tell().
        Returns:
            True if any of the stopping conditions are satisfied.
        """

    @abstractmethod
    def ask(self, lower_bounds, upper_bounds):
        """Samples new solutions from the Gaussian distribution.

        Args:
            lower_bounds (float or np.ndarray): scalar or (solution_dim,) array
                indicating lower bounds of the solution space. Scalars specify
                the same bound for the entire space, while arrays specify a
                bound for each dimension. Pass -np.inf in the array or scalar to
                indicated unbounded space.
            upper_bounds (float or np.ndarray): Same as above, but for upper
                bounds (and pass np.inf instead of -np.inf).
        """

    @abstractmethod
    def tell(self, ranking_indices, num_parents):
        """Passes the solutions back to the optimizer.

        Args:
            ranking_indices (array-like of int): Indices that indicate the
                ranking of the original solutions returned in ``ask()``.
            num_parents (int): Number of top solutions to select from the
                ranked solutions.
        """



"""Provides EmitterBase."""
from abc import ABC

import numpy as np


class EmitterBase(ABC):
    """Base class for emitters.

    Every emitter has an :meth:`ask` method that generates a batch of solutions,
    and a :meth:`tell` method that inserts solutions into the emitter's archive.
    Child classes are only required to override :meth:`ask`.

    DQD emitter should override :meth:`ask_dqd` and :meth:`tell_dqd` methods.

    Args:
        archive (ribs.archives.ArchiveBase): An archive to use when creating and
            inserting solutions. For instance, this can be
            :class:`ribs.archives.GridArchive`.
        solution_dim (int): The dimension of solutions produced by this emitter.
        bounds (None or array-like): Bounds of the solution space. Pass None to
            indicate there are no bounds. Alternatively, pass an array-like to
            specify the bounds for each dim. Each element in this array-like can
            be None to indicate no bound, or a tuple of
            ``(lower_bound, upper_bound)``, where ``lower_bound`` or
            ``upper_bound`` may be None to indicate no bound.

            Unbounded upper bounds are set to +inf, and unbounded lower bounds
            are set to -inf.
    """

    def __init__(self, archive, *, solution_dim, bounds):
        self._archive = archive
        self._solution_dim = solution_dim
        (self._lower_bounds,
         self._upper_bounds) = self._process_bounds(bounds, self._solution_dim,
                                                    archive.dtype)

    @staticmethod
    def _process_bounds(bounds, solution_dim, dtype):
        """Processes the input bounds.

        Returns:
            tuple: Two arrays containing all the lower bounds and all the upper
                bounds.
        Raises:
            ValueError: There is an error in the bounds configuration.
        """
        lower_bounds = np.full(solution_dim, -np.inf, dtype=dtype)
        upper_bounds = np.full(solution_dim, np.inf, dtype=dtype)

        if bounds is None:
            return lower_bounds, upper_bounds

        # Handle array-like bounds.
        if len(bounds) != solution_dim:
            raise ValueError("If it is an array-like, bounds must have the "
                             "same length as x0")
        for idx, bnd in enumerate(bounds):
            if bnd is None:
                continue  # Bounds already default to -inf and inf.
            if len(bnd) != 2:
                raise ValueError("All entries of bounds must be length 2")
            lower_bounds[idx] = -np.inf if bnd[0] is None else bnd[0]
            upper_bounds[idx] = np.inf if bnd[1] is None else bnd[1]
        return lower_bounds, upper_bounds

    @property
    def archive(self):
        """ribs.archives.ArchiveBase: The archive which stores solutions
        generated by this emitter."""
        return self._archive

    @property
    def solution_dim(self):
        """int: The dimension of solutions produced by this emitter."""
        return self._solution_dim

    @property
    def lower_bounds(self):
        """numpy.ndarray: ``(solution_dim,)`` array with lower bounds of
        solution space.

        For instance, ``[-1, -1, -1]`` indicates that every dimension of the
        solution space has a lower bound of -1.
        """
        return self._lower_bounds

    @property
    def upper_bounds(self):
        """numpy.ndarray: ``(solution_dim,)`` array with upper bounds of
        solution space.

        For instance, ``[1, 1, 1]`` indicates that every dimension of the
        solution space has an upper bound of 1.
        """
        return self._upper_bounds

    def ask(self):
        """Generates a ``(batch_size, solution_dim)`` array of solutions.

        Returns an empty array by default.
        """
        return np.empty((0, self.solution_dim), dtype=self.archive.dtype)

    def tell(self,
             solution_batch,
             objective_batch,
             measures_batch,
             status_batch,
             value_batch,
             metadata_batch=None):
        """Gives the emitter results from evaluating solutions.

        This base class implementation (in :class:`~ribs.emitters.EmitterBase`)
        needs to be overriden.

        Args:
            solution_batch (numpy.ndarray): Array of solutions generated by this
                emitter's :meth:`ask()` method.
            objective_batch (numpy.ndarray): 1D array containing the objective
                function value of each solution.
            measures_batch (numpy.ndarray): ``(n, <measure space dimension>)``
                array with the measure space coordinates of each solution.
            status_batch (numpy.ndarray): An array of integer statuses
                returned by a series of calls to archive's :meth:`add_single()`
                method or by a single call to archive's :meth:`add()`.
            value_batch  (numpy.ndarray): 1D array of floats returned by a
                series of calls to archive's :meth:`add_single()` method or by a
                single call to archive's :meth:`add()`. For what these floats
                represent, refer to :meth:`ribs.archives.add()`.
            metadata_batch (numpy.ndarray): 1D object array containing a
                metadata object for each solution.
        """

    def ask_dqd(self):
        """Generates a ``(batch_size, solution_dim)`` array of solutions for
        which gradient information must be computed.

        This method only needs to be implemented by emitters used in DQD. The
        method returns an empty array by default.
        """
        return np.empty((0, self.solution_dim), dtype=self.archive.dtype)

    def tell_dqd(self,
                 solution_batch,
                 objective_batch,
                 measures_batch,
                 jacobian_batch,
                 status_batch,
                 value_batch,
                 metadata_batch=None):
        """Gives the emitter results from evaluating the gradient of the
        solutions, only used for DQD emitters.

        Args:
            solution_batch (numpy.ndarray): ``(batch_size,
                :attr:`solution_dim`)`` array of solutions generated by this
                emitter's :meth:`ask()` method.
            objective_batch (numpy.ndarray): 1-dimensional array containing the
                objective function value of each solution.
            measures_batch (numpy.ndarray): ``(batch_size, measure space
                dimension)`` array with the measure space coordinates of each
                solution.
            jacobian_batch (numpy.ndarray): ``(batch_size, 1 + measure_dim,
                solution_dim)`` array consisting of Jacobian matrices of the
                solutions obtained from :meth:`ask_dqd`. Each matrix should
                consist of the objective gradient of the solution followed by
                the measure gradients.
            status_batch (numpy.ndarray): 1d array of
                :class:`ribs.archive.addstatus` returned by a series of calls
                to archive's :meth:`add()` method.
            value_batch (numpy.ndarray): 1d array of floats returned by a series
                of calls to archive's :meth:`add()` method. for what these
                floats represent, refer to :meth:`ribs.archives.add()`.
            metadata_batch (numpy.ndarray): 1d object array containing a
                metadata object for each solution.
        """



"""Rankers for use across emitters.

The rankers implemented in this file are intended to be used with emitters.
Specifically, a ranker object should be initialized or passed in the emitters.
The ``Ranker`` object will define the :meth:`rank` method which returns the
result of a descending argsort of the solutions. It will also define a
:meth:`reset` method which resets the internal state of the object.

When specifying which ranker to use for each emitter, one could either pass in
the full name of a ranker, e.g. "ImprovementRanker", or the abbreviated name of
a ranker, e.g. "imp".
The supported abbreviations are:

* ``imp``: :class:`ImprovementRanker`
* ``2imp``: :class:`TwoStageImprovementRanker`
* ``rd``: :class:`RandomDirectionRanker`
* ``2rd``: :class:`TwoStageRandomDirectionRanker`
* ``obj``: :class:`ObjectiveRanker`
* ``2obj``: :class:`TwoStageObjectiveRanker`

.. autosummary::
    :toctree:

    ribs.emitters.rankers.ImprovementRanker
    ribs.emitters.rankers.TwoStageImprovementRanker
    ribs.emitters.rankers.RandomDirectionRanker
    ribs.emitters.rankers.TwoStageRandomDirectionRanker
    ribs.emitters.rankers.ObjectiveRanker
    ribs.emitters.rankers.TwoStageObjectiveRanker
    ribs.emitters.rankers.RankerBase
"""
from abc import ABC, abstractmethod

import numpy as np

from ribs._docstrings import DocstringComponents, core_args

__all__ = [
    "ImprovementRanker",
    "TwoStageImprovementRanker",
    "RandomDirectionRanker",
    "TwoStageRandomDirectionRanker",
    "ObjectiveRanker",
    "TwoStageObjectiveRanker",
    "RankerBase",
]

# Define common docstrings
_args = DocstringComponents(core_args)

_rank_args = f"""
Args:
    emitter (ribs.emitters.EmitterBase): Emitter that this ``ranker``
        object belongs to.
    archive (ribs.archives.ArchiveBase): Archive used by ``emitter``
        when creating and inserting solutions.
    rng (numpy.random.Generator): A random number generator.
{_args.solution_batch}
{_args.objective_batch}
{_args.measures_batch}
{_args.status_batch}
{_args.value_batch}
{_args.metadata_batch}

Returns:
    tuple(numpy.ndarray, numpy.ndarray): the first array (shape
    ``(batch_size,)``) is an array of indices representing a ranking of the
    solutions and the second array (shape ``(batch_size,)`` or (batch_size,
    n_values)``) is an array of values that this ranker used to rank the
    solutions. ``batch_size`` is the number of solutions and ``n_values`` is
    the number of values that the rank function used.
"""

_reset_args = """
Args:
    emitter (ribs.emitters.EmitterBase): Emitter that this ``ranker``
        object belongs to.
    archive (ribs.archives.ArchiveBase): Archive used by ``emitter``
        when creating and inserting solutions.
    rng (numpy.random.Generator): A random number generator.
"""


class RankerBase(ABC):
    """Base class for rankers.

    Every ranker has a :meth:`rank` method that returns a list of indices
    that indicate how the solutions should be ranked and a :meth:`reset` method
    that resets the internal state of the ranker
    (e.g. in :class:`ribs.emitters.rankers._random_direction_ranker`).

    Child classes are only required to override :meth:`rank`.
    """

    @abstractmethod
    def rank(self, emitter, archive, rng, solution_batch, objective_batch,
             measures_batch, status_batch, value_batch, metadata_batch):
        # pylint: disable=missing-function-docstring
        pass

    # Generates the docstring for rank
    rank.__doc__ = f"""
Generates a batch of indices that represents an ordering of ``solution_batch``.

{_rank_args}
    """

    def reset(self, emitter, archive, rng):
        # pylint: disable=missing-function-docstring
        pass

    reset.__doc__ = f"""
Resets the internal state of the ranker.

{_reset_args}
   """


class ImprovementRanker(RankerBase):
    """Ranks the solutions based on the improvement in the objective.

    This ranker ranks solutions in a single stage. The solutions are ranked by
    the improvement "value" described in :meth:`ribs.archives.ArchiveBase.add`.

    This ranker should not be used with CMA-ME. The improvement values for new
    solutions in CMA-ME are on a different scale from the improvement values for
    the other solutions, in that new solutions have an improvement value which
    is simply their objective, while other solutions have an improvement value
    which is the difference between their objective and the objective of their
    corresponding cell in the archive.
    """

    def rank(self, emitter, archive, rng, solution_batch, objective_batch,
             measures_batch, status_batch, value_batch, metadata_batch):
        # Note that lexsort sorts the values in ascending order,
        # so we use np.flip to reverse the sorted array.
        return np.flip(np.argsort(value_batch)), value_batch

    rank.__doc__ = f"""
Generates a list of indices that represents an ordering of solutions.

{_rank_args}
    """


class TwoStageImprovementRanker(RankerBase):
    """Ranks the solutions based on the improvement in the objective.

    This ranker originates in `Fontaine 2020
    <https://arxiv.org/abs/1912.02400>`_ in which it was referred to as
    "Improvement Emitter".
    This ranker ranks solutions in two stages. First, solutions are ranked by
    "status" -- those that found a new cell in the archive rank above those that
    improved an existing cell, which rank above those that were not added to the
    archive. Second, solutions are ranked by the "value" described in
    :meth:`ribs.archives.ArchiveBase.add`
    """

    def rank(self, emitter, archive, rng, solution_batch, objective_batch,
             measures_batch, status_batch, value_batch, metadata_batch):
        # To avoid using an array of tuples, ranking_values is a 2D array
        # [[status_0, value_0], ..., [status_n, value_n]]
        ranking_values = np.stack((status_batch, value_batch), axis=-1)

        # New solutions sort ahead of improved ones, which sort ahead of ones
        # that were not added.
        #
        # Since lexsort uses the last column/row as the key, we flip the
        # ranking_values along the last axis so that we are sorting by status
        # first.
        #
        # Note that lexsort sorts the values in ascending order,
        # so we use np.flip to reverse the sorted array.
        return (
            np.flip(np.lexsort(np.flip(ranking_values, axis=-1).T)),
            ranking_values,
        )

    rank.__doc__ = f"""
Generates a list of indices that represents an ordering of solutions.

{_rank_args}
    """


class RandomDirectionRanker(RankerBase):
    """Ranks the solutions based on projection onto a direction in measure
    space.

    This ranker originates in `Fontaine 2020
    <https://arxiv.org/abs/1912.02400>`_ as random direction emitter.
    The solutions are ranked solely based on their projection onto a random
    direction in the archive.

    To rank the solutions first by whether they were added, and then by
    the projection, refer to
    :class:`ribs.emitters.rankers.TwoStageRandomDirectionRanker`.
    """

    def __init__(self):
        super().__init__()
        self._target_measure_dir = None

    @property
    def target_measure_dir(self):
        """numpy.ndarray: ``(measure_dim,)`` array with the target measure
        direction vector."""
        return self._target_measure_dir

    @target_measure_dir.setter
    def target_measure_dir(self, value):
        self._target_measure_dir = value

    def rank(self, emitter, archive, rng, solution_batch, objective_batch,
             measures_batch, status_batch, value_batch, metadata_batch):
        if self._target_measure_dir is None:
            raise RuntimeError("target measure direction not set")
        projections = np.dot(measures_batch, self._target_measure_dir)
        # Sort only by projection; use np.flip to reverse the order
        return np.flip(np.argsort(projections)), projections

    rank.__doc__ = f"""
Ranks the solutions based on projection onto a direction in the archive.

{_rank_args}
    """

    def reset(self, emitter, archive, rng):
        ranges = archive.upper_bounds - archive.lower_bounds
        measure_dim = len(ranges)
        unscaled_dir = rng.standard_normal(measure_dim)
        self._target_measure_dir = unscaled_dir * ranges

    # Generates the docstring for reset.
    reset.__doc__ = f"""
Generates a random direction in the archive's measure space.

A random direction is sampled from a standard Gaussian -- since the standard
Gaussian is isotropic, there is equal probability for any direction. The
direction is then scaled to the archive bounds so that it is a random archive
direction.

{_reset_args}
   """


class TwoStageRandomDirectionRanker(RankerBase):
    """Similar to :class:`ribs.emitters.rankers.RandomDirectionRanker`, but the
    solutions are first ranked by whether they are added, then by their
    projection onto a random direction in the archive space.

    This ranker originates from `Fontaine 2020
    <https://arxiv.org/abs/1912.02400>`_ as RandomDirectionEmitter.
    """

    def __init__(self):
        super().__init__()
        self._target_measure_dir = None

    @property
    def target_measure_dir(self):
        """numpy.ndarray: ``(measure_dim,)`` array with the target measure
        direction vector."""
        return self._target_measure_dir

    @target_measure_dir.setter
    def target_measure_dir(self, value):
        self._target_measure_dir = value

    def rank(self, emitter, archive, rng, solution_batch, objective_batch,
             measures_batch, status_batch, value_batch, metadata_batch):
        if self._target_measure_dir is None:
            raise RuntimeError("target measure direction not set")
        projections = np.dot(measures_batch, self._target_measure_dir)

        # To avoid using an array of tuples, ranking_values is a 2D array
        # [[status_0, projection_0], ..., [status_n, projection_n]]
        ranking_values = np.stack((status_batch, projections), axis=-1)

        # Sort by whether the solution was added into the archive,
        # followed by projection.
        return (
            np.flip(np.lexsort(np.flip(ranking_values, axis=-1).T)),
            ranking_values,
        )

    rank.__doc__ = f"""
Ranks the solutions first by whether they are added, then by their projection
onto a random direction in the archive.

{_rank_args}
    """

    def reset(self, emitter, archive, rng):
        ranges = archive.upper_bounds - archive.lower_bounds
        measure_dim = len(ranges)
        unscaled_dir = rng.standard_normal(measure_dim)
        self._target_measure_dir = unscaled_dir * ranges

    reset.__doc__ = RandomDirectionRanker.reset.__doc__


class ObjectiveRanker(RankerBase):
    """Ranks the solutions solely based on the objective values.

    This ranker originates in `Fontaine 2020
    <https://arxiv.org/abs/1912.02400>`_ in which it was part of the optimizing
    emitter.
    """

    def rank(self, emitter, archive, rng, solution_batch, objective_batch,
             measures_batch, status_batch, value_batch, metadata_batch):
        # Sort only by objective value.
        return np.flip(np.argsort(objective_batch)), objective_batch

    rank.__doc__ = f"""
Ranks the solutions based on their objective values.

{_rank_args}
    """


class TwoStageObjectiveRanker(RankerBase):
    """Similar to :class:`ribs.emitters.rankers.ObjectiveRanker`, but ranks
    newly added solutions before improved solutions.

    This ranker originates in `Fontaine 2020
    <https://arxiv.org/abs/1912.02400>`_ as OptimizingEmitter.
    """

    def rank(self, emitter, archive, rng, solution_batch, objective_batch,
             measures_batch, status_batch, value_batch, metadata_batch):
        # To avoid using an array of tuples, ranking_values is a 2D array
        # [[status_0, objective_0], ..., [status_0, objective_n]]
        ranking_values = np.stack((status_batch, objective_batch), axis=-1)

        # Sort by whether the solution was added into the archive, followed
        # by the objective values.
        return (
            np.flip(np.lexsort(np.flip(ranking_values, axis=-1).T)),
            ranking_values,
        )

    rank.__doc__ = f"""
Ranks the solutions based on their objective values, while prioritizing newly
added solutions.

{_rank_args}
    """


_NAME_TO_RANKER_MAP = {
    "ImprovementRanker": ImprovementRanker,
    "TwoStageImprovementRanker": TwoStageImprovementRanker,
    "RandomDirectionRanker": RandomDirectionRanker,
    "TwoStageRandomDirectionRanker": TwoStageRandomDirectionRanker,
    "ObjectiveRanker": ObjectiveRanker,
    "TwoStageObjectiveRanker": TwoStageObjectiveRanker,
    "imp": ImprovementRanker,
    "2imp": TwoStageImprovementRanker,
    "rd": RandomDirectionRanker,
    "2rd": TwoStageRandomDirectionRanker,
    "obj": ObjectiveRanker,
    "2obj": TwoStageObjectiveRanker
}


def _get_ranker(klass):
    """Returns a ranker class based on its name.

    ``klass`` can be a reference to the class of the ranker, the full name of
    a ranker, e.g. "ImprovementRanker", or the abbreviated name for a ranker
    such as "imp".

    Args:
        klass (callable or str): This parameter may either be a callable (e.g.
            a class or a lambda function) that takes in no parameters and
            returns an instance of :class:`RankerBase`, or it may be a full or
            abbreviated ranker name.

    Returns:
        The corresponding ranker class instance.
    """
    if isinstance(klass, str):
        if klass in _NAME_TO_RANKER_MAP:
            return _NAME_TO_RANKER_MAP[klass]()
        raise ValueError(f"`{klass}` is not the full or abbreviated "
                         "name of a valid ranker")
    if callable(klass):
        ranker = klass()
        if isinstance(ranker, RankerBase):
            return ranker
        raise ValueError(f"Callable `{klass}` did not return an instance "
                         "of RankerBase.")
    raise ValueError(f"`{klass}` is neither a callable nor a string")



"""Contains the GridArchive."""
import numpy as np

from ribs._utils import check_batch_shape, check_finite, check_is_1d
from ribs.archives._archive_base import ArchiveBase


class GridArchive(ArchiveBase):
    """An archive that divides each dimension into uniformly-sized cells.

    This archive is the container described in `Mouret 2015
    <https://arxiv.org/pdf/1504.04909.pdf>`_. It can be visualized as an
    n-dimensional grid in the measure space that is divided into a certain
    number of cells in each dimension. Each cell contains an elite, i.e. a
    solution that `maximizes` the objective function for the measures in that
    cell.

    .. note:: The idea of archive thresholds was introduced in `Fontaine 2022
        <https://arxiv.org/abs/2205.10752>`_. Refer to our `CMA-MAE tutorial
        <../../tutorials/cma_mae.html>`_ for more info on thresholds, including
        the ``learning_rate`` and ``threshold_min`` parameters.

    Args:
        solution_dim (int): Dimension of the solution space.
        dims (array-like of int): Number of cells in each dimension of the
            measure space, e.g. ``[20, 30, 40]`` indicates there should be 3
            dimensions with 20, 30, and 40 cells. (The number of dimensions is
            implicitly defined in the length of this argument).
        ranges (array-like of (float, float)): Upper and lower bound of each
            dimension of the measure space, e.g. ``[(-1, 1), (-2, 2)]``
            indicates the first dimension should have bounds :math:`[-1,1]`
            (inclusive), and the second dimension should have bounds
            :math:`[-2,2]` (inclusive). ``ranges`` should be the same length as
            ``dims``.
        epsilon (float): Due to floating point precision errors, we add a small
            epsilon when computing the archive indices in the :meth:`index_of`
            method -- refer to the implementation `here
            <../_modules/ribs/archives/_grid_archive.html#GridArchive.index_of>`_.
            Pass this parameter to configure that epsilon.
        learning_rate (float): The learning rate for threshold updates.
        threshold_min (float): The initial threshold value for all the cells.
        qd_score_offset (float): Archives often contain negative objective
            values, and if the QD score were to be computed with these negative
            objectives, the algorithm would be penalized for adding new cells
            with negative objectives. Thus, a standard practice is to normalize
            all the objectives so that they are non-negative by introducing an
            offset. This QD score offset will be *subtracted* from all
            objectives in the archive, e.g., if your objectives go as low as
            -300, pass in -300 so that each objective will be transformed as
            ``objective - (-300)``.
        seed (int): Value to seed the random number generator. Set to None to
            avoid a fixed seed.
        dtype (str or data-type): Data type of the solutions, objectives,
            and measures. We only support ``"f"`` / ``np.float32`` and ``"d"`` /
            ``np.float64``.
    Raises:
        ValueError: ``dims`` and ``ranges`` are not the same length.
    """

    def __init__(self,
                 *,
                 solution_dim,
                 dims,
                 ranges,
                 learning_rate=1.0,
                 threshold_min=-np.inf,
                 epsilon=1e-6,
                 qd_score_offset=0.0,
                 seed=None,
                 dtype=np.float64):
        self._dims = np.array(dims, dtype=np.int32)
        if len(self._dims) != len(ranges):
            raise ValueError(f"dims (length {len(self._dims)}) and ranges "
                             f"(length {len(ranges)}) must be the same length")

        ArchiveBase.__init__(
            self,
            solution_dim=solution_dim,
            cells=np.product(self._dims),
            measure_dim=len(self._dims),
            learning_rate=learning_rate,
            threshold_min=threshold_min,
            qd_score_offset=qd_score_offset,
            seed=seed,
            dtype=dtype,
        )

        ranges = list(zip(*ranges))
        self._lower_bounds = np.array(ranges[0], dtype=self.dtype)
        self._upper_bounds = np.array(ranges[1], dtype=self.dtype)
        self._interval_size = self._upper_bounds - self._lower_bounds
        self._epsilon = self.dtype(epsilon)

        self._boundaries = []
        for dim, lower_bound, upper_bound in zip(self._dims, self._lower_bounds,
                                                 self._upper_bounds):
            self._boundaries.append(
                np.linspace(lower_bound, upper_bound, dim + 1))

    @property
    def dims(self):
        """(measure_dim,) numpy.ndarray: Number of cells in each dimension."""
        return self._dims

    @property
    def lower_bounds(self):
        """(measure_dim,) numpy.ndarray: Lower bound of each dimension."""
        return self._lower_bounds

    @property
    def upper_bounds(self):
        """(measure_dim,) numpy.ndarray: Upper bound of each dimension."""
        return self._upper_bounds

    @property
    def interval_size(self):
        """(measure_dim,) numpy.ndarray: The size of each dim (upper_bounds -
        lower_bounds)."""
        return self._interval_size

    @property
    def epsilon(self):
        """:attr:`dtype`: Epsilon for computing archive indices. Refer to
        the documentation for this class."""
        return self._epsilon

    @property
    def boundaries(self):
        """list of numpy.ndarray: The boundaries of the cells in each dimension.

        Entry ``i`` in this list is an array that contains the boundaries of the
        cells in dimension ``i``. The array contains ``self.dims[i] + 1``
        entries laid out like this::

            Archive cells:  | 0 | 1 |   ...   |    self.dims[i]    |
            boundaries[i]:  0   1   2   self.dims[i] - 1     self.dims[i]

        Thus, ``boundaries[i][j]`` and ``boundaries[i][j + 1]`` are the lower
        and upper bounds of cell ``j`` in dimension ``i``. To access the lower
        bounds of all the cells in dimension ``i``, use ``boundaries[i][:-1]``,
        and to access all the upper bounds, use ``boundaries[i][1:]``.
        """
        return self._boundaries

    def index_of(self, measures_batch):
        """Returns archive indices for the given batch of measures.

        First, values are clipped to the bounds of the measure space. Then, the
        values are mapped to cells; e.g. cell 5 along dimension 0 and cell 3
        along dimension 1.

        At this point, we have "grid indices" -- indices of each measure in each
        dimension. Since indices returned by this method must be single integers
        (as opposed to a tuple of grid indices), we convert these grid indices
        into integer indices with :func:`numpy.ravel_multi_index` and return the
        result.

        It may be useful to have the original grid indices. Thus, we provide the
        :meth:`grid_to_int_index` and :meth:`int_to_grid_index` methods for
        converting between grid and integer indices.

        As an example, the grid indices can be used to access boundaries of a
        measure value's cell. For example, the following retrieves the lower
        and upper bounds of the cell along dimension 0::

            # Access only element 0 since this method operates in batch.
            idx = archive.int_to_grid_index(archive.index_of(...))[0]

            lower = archive.boundaries[0][idx[0]]
            upper = archive.boundaries[0][idx[0] + 1]

        See :attr:`boundaries` for more info.

        Args:
            measures_batch (array-like): (batch_size, :attr:`measure_dim`)
                array of coordinates in measure space.
        Returns:
            numpy.ndarray: (batch_size,) array of integer indices representing
            the flattened grid coordinates.
        Raises:
            ValueError: ``measures_batch`` is not of shape (batch_size,
                :attr:`measure_dim`).
            ValueError: ``measures_batch`` has non-finite values (inf or NaN).
        """
        measures_batch = np.asarray(measures_batch)
        check_batch_shape(measures_batch, "measures_batch", self.measure_dim,
                          "measure_dim")
        check_finite(measures_batch, "measures_batch")

        # Adding epsilon accounts for floating point precision errors from
        # transforming measures. We then cast to int32 to obtain integer
        # indices.
        grid_index_batch = (
            (self._dims *
             (measures_batch - self._lower_bounds) + self._epsilon) /
            self._interval_size).astype(np.int32)

        # Clip indices to the archive dimensions (for example, for 20 cells, we
        # want indices to run from 0 to 19).
        grid_index_batch = np.clip(grid_index_batch, 0, self._dims - 1)

        return self.grid_to_int_index(grid_index_batch)

    def grid_to_int_index(self, grid_index_batch):
        """Converts a batch of grid indices into a batch of integer indices.

        Refer to :meth:`index_of` for more info.

        Args:
            grid_index_batch (array-like): (batch_size, :attr:`measure_dim`)
                array of indices in the archive grid.
        Returns:
            numpy.ndarray: (batch_size,) array of integer indices.
        Raises:
            ValueError: ``grid_index_batch`` is not of shape (batch_size,
                :attr:`measure_dim`)
        """
        grid_index_batch = np.asarray(grid_index_batch)
        check_batch_shape(grid_index_batch, "grid_index_batch",
                          self.measure_dim, "measure_dim")

        return np.ravel_multi_index(grid_index_batch.T,
                                    self._dims).astype(np.int32)

    def int_to_grid_index(self, int_index_batch):
        """Converts a batch of indices into indices in the archive's grid.

        Refer to :meth:`index_of` for more info.

        Args:
            int_index_batch (array-like): (batch_size,) array of integer
                indices such as those output by :meth:`index_of`.
        Returns:
            numpy.ndarray: (batch_size, :attr:`measure_dim`) array of indices
            in the archive grid.
        Raises:
            ValueError: ``int_index_batch`` is not of shape (batch_size,).
        """
        int_index_batch = np.asarray(int_index_batch)
        check_is_1d(int_index_batch, "int_index_batch")

        return np.asarray(np.unravel_index(
            int_index_batch,
            self._dims,
        )).T.astype(np.int32)



"""This provides the common docstrings that are used throughout the project."""

import re


class DocstringComponents:
    """Adapted from https://github.com/mwaskom/seaborn/blob/9d8ce6ad4ab213994f0b
    c84d0c46869df7be0b49/seaborn/_docstrings.py."""
    regexp = re.compile(r"\n((\n|.)+)\n\s*", re.MULTILINE)

    def __init__(self, comp_dict, strip_whitespace=True):
        """Read entries from a dict, optionally stripping outer whitespace."""
        if strip_whitespace:
            entries = {}
            for key, val in comp_dict.items():
                m = re.match(self.regexp, val)
                if m is None:
                    entries[key] = val
                else:
                    entries[key] = m.group(1)
        else:
            entries = comp_dict.copy()

        self.entries = entries

    def __getattr__(self, attr):
        """Provide dot access to entries for clean raw docstrings."""
        if attr in self.entries:
            return self.entries[attr]
        try:
            return self.__getattribute__(attr)
        except AttributeError as err:
            # If Python is run with -OO, it will strip docstrings and our lookup
            # from self.entries will fail. We check for __debug__, which is actually
            # set to False by -O (it is True for normal execution).
            # But we only want to see an error when building the docs;
            # not something users should see, so this slight inconsistency is fine.
            if __debug__:
                raise err
            else:
                pass

    @classmethod
    def from_nested_components(cls, **kwargs):
        """Add multiple sub-sets of components."""
        return cls(kwargs, strip_whitespace=False)

    # NOTE Unclear how this will be useful, commenting out for now
    # @classmethod
    # def from_function_params(cls, func):
    #     """Use the numpydoc parser to extract components from existing func."""
    #     params = NumpyDocString(pydoc.getdoc(func))["Parameters"]
    #     comp_dict = {}
    #     for p in params:
    #         name = p.name
    #         type = p.type
    #         desc = "\n    ".join(p.desc)
    #         comp_dict[name] = f"{name} : {type}\n    {desc}"

    #     return cls(comp_dict)


core_args = dict(emitter="""
    emitter (ribs.emitters.EmitterBase): Emitter to use for generating
        solutions and updating the archive.
    """,
                 archive="""
    archive (ribs.archives.ArchiveBase): Archive to use when creating
        and inserting solutions. For instance, this can be
        :class:`ribs.archives.GridArchive`.
    """,
                 solution_batch="""
    solution_batch (numpy.ndarray): Batch of solutions generated by the
        emitter's :meth:`ask()` method.
    """,
                 objective_batch="""
    objective_batch (numpy.ndarray): Batch of objective values.
    """,
                 measures_batch="""
    measures_batch (numpy.ndarray): ``(n, <measure space dimension>)``
        array with the measure space coordinates of each solution.
    """,
                 metadata_batch="""
    metadata_batch (numpy.ndarray): 1D object array containing a metadata
        object for each solution.
    """,
                 status_batch="""
    status_batch (numpy.ndarray): An array of integer statuses
        returned by a series of calls to archive's :meth:`add_single()`
        method or by a single call to archive's :meth:`add()`.
    """,
                 value_batch="""
    value_batch  (numpy.ndarray): 1D array of floats returned by a series of
        calls to archive's :meth:`add_single()` method or by a single call to
        archive's :meth:`add()`. For what these floats represent,
        refer to :meth:`ribs.archives.add()`.
    """,
                 seed="""
    seed (int): Value to seed the random number generator. Set to None to
        avoid a fixed seed.
    """)

core_docs = dict(args=DocstringComponents(core_args))



#!/usr/bin/env python
"""The setup script."""

from setuptools import find_packages, setup

with open("README.md") as readme_file:
    readme = readme_file.read()

with open("HISTORY.md") as history_file:
    history = history_file.read()

# NOTE: Update pinned_reqs whenever install_requires or extras_require changes.
install_requires = [
    "numpy>=1.17.0",  # >=1.17.0 is when default_rng becomes available.
    "numpy_groupies>=0.9.16",  # Supports Python 3.7 and up.
    "numba>=0.51.0",
    "pandas>=1.0.0",
    "sortedcontainers>=2.0.0",  # Primarily used in SlidingBoundariesArchive.
    "scikit-learn>=0.20.0",  # Primarily used in CVTArchive.
    "scipy>=1.4.0",  # Primarily used in CVTArchive.
    "threadpoolctl>=3.0.0",
    "semantic-version>=2.10"
]

extras_require = {
    "visualize": ["matplotlib>=3.0.0",],
    # All dependencies except for dev. Don't worry if there are duplicate
    # dependencies, since setuptools automatically handles duplicates.
    "all": [
        ### visualize ###
        "matplotlib>=3.0.0",
    ],
    "dev": [
        "pip>=20.3",
        "pylint==2.8.3",
        "yapf",

        # Testing
        "pytest==7.0.1",
        "pytest-cov==3.0.0",
        "pytest-benchmark==3.4.1",
        "pytest-xdist==2.5.0",

        # Documentation
        "Sphinx==4.5.0",
        "sphinx-material==0.0.32",
        "sphinx-autobuild==2021.3.14",
        "sphinx-copybutton==0.3.1",
        "myst-nb==0.17.1",
        "sphinx-toolbox==3.1.0",
        "sphinx-autodoc-typehints==1.18.2",
        "sphinx-codeautolink==0.12.1",

        # Distribution
        "bump2version==0.5.11",
        "wheel==0.36.2",
        "twine==1.14.0",
        "check-wheel-contents==0.2.0",
    ],
}

setup(
    author="ICAROS Lab pyribs Team",
    author_email="team@pyribs.org",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    description=
    "A bare-bones Python library for quality diversity optimization.",
    install_requires=install_requires,
    extras_require=extras_require,
    license="MIT license",
    long_description=readme + "\n\n" + history,
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords="ribs",
    name="ribs",
    packages=find_packages(include=["ribs", "ribs.*"]),
    python_requires=">=3.7.0",
    test_suite="tests",
    url="https://github.com/icaros-usc/pyribs",
    version="0.5.1",
    zip_safe=False,
)



"""Schedulers link the entire QD algorithm together.

.. autosummary::
    :toctree:

    ribs.schedulers.Scheduler
    ribs.schedulers.BanditScheduler
"""
from ribs.schedulers._scheduler import Scheduler
from ribs.schedulers._bandit_scheduler import BanditScheduler

__all__ = [
    "Scheduler",
    "BanditScheduler",
]



"""Provides the GaussianEmitter."""
import numpy as np

from ribs._utils import check_1d_shape, check_batch_shape
from ribs.emitters._emitter_base import EmitterBase


class GaussianEmitter(EmitterBase):
    """Emits solutions by adding Gaussian noise to existing archive solutions.

    If the archive is empty and ``self._initial_solutions`` is set, a call to
    :meth:`ask` will return ``self._initial_solutions``. If
    ``self._initial_solutions`` is not set, we draw from a Gaussian distribution
    centered at ``self.x0`` with standard deviation ``self.sigma``. Otherwise,
    each solution is drawn from a distribution centered at a randomly chosen
    elite with standard deviation ``self.sigma``.

    This is the classic variation operator presented in `Mouret 2015
    <https://arxiv.org/pdf/1504.04909.pdf>`_.

    Args:
        archive (ribs.archives.ArchiveBase): An archive to use when creating and
            inserting solutions. For instance, this can be
            :class:`ribs.archives.GridArchive`.
        sigma (float or array-like): Standard deviation of the Gaussian
            distribution. Note we assume the Gaussian is diagonal, so if this
            argument is an array, it must be 1D.
        x0 (array-like): Center of the Gaussian distribution from which to
            sample solutions when the archive is empty. Must be 1-dimensional.
            This argument is ignored if ``initial_solutions`` is set.
        initial_solutions (array-like): An (n, solution_dim) array of solutions
            to be used when the archive is empty. If this argument is None, then
            solutions will be sampled from a Gaussian distribution centered at
            ``x0`` with standard deviation ``sigma``.
        bounds (None or array-like): Bounds of the solution space. Solutions are
            clipped to these bounds. Pass None to indicate there are no bounds.
            Alternatively, pass an array-like to specify the bounds for each
            dim. Each element in this array-like can be None to indicate no
            bound, or a tuple of ``(lower_bound, upper_bound)``, where
            ``lower_bound`` or ``upper_bound`` may be None to indicate no bound.
        batch_size (int): Number of solutions to return in :meth:`ask`.
        seed (int): Value to seed the random number generator. Set to None to
            avoid a fixed seed.
    Raises:
        ValueError: There is an error in x0 or initial_solutions.
        ValueError: There is an error in the bounds configuration.
    """

    def __init__(self,
                 archive,
                 *,
                 sigma,
                 x0=None,
                 initial_solutions=None,
                 bounds=None,
                 batch_size=64,
                 seed=None):
        self._rng = np.random.default_rng(seed)
        self._batch_size = batch_size

        self._sigma = np.array(sigma, dtype=archive.dtype)

        self._x0 = None
        self._initial_solutions = None

        if x0 is None and initial_solutions is None:
            raise ValueError("Either x0 or initial_solutions must be provided.")
        if x0 is not None and initial_solutions is not None:
            raise ValueError(
                "x0 and initial_solutions cannot both be provided.")

        if x0 is not None:
            self._x0 = np.array(x0, dtype=archive.dtype)
            check_1d_shape(self._x0, "x0", archive.solution_dim,
                           "archive.solution_dim")
        elif initial_solutions is not None:
            self._initial_solutions = np.asarray(initial_solutions,
                                                 dtype=archive.dtype)
            check_batch_shape(self._initial_solutions, "initial_solutions",
                              archive.solution_dim, "archive.solution_dim")

        EmitterBase.__init__(
            self,
            archive,
            solution_dim=archive.solution_dim,
            bounds=bounds,
        )

    @property
    def x0(self):
        """numpy.ndarray: Center of the Gaussian distribution from which to
        sample solutions when the archive is empty (if initial_solutions is not
        set)."""
        return self._x0

    @property
    def initial_solutions(self):
        """numpy.ndarray: The initial solutions which are returned when the
        archive is empty (if x0 is not set)."""
        return self._initial_solutions

    @property
    def sigma(self):
        """float or numpy.ndarray: Standard deviation of the (diagonal) Gaussian
        distribution when the archive is not empty."""
        return self._sigma

    @property
    def batch_size(self):
        """int: Number of solutions to return in :meth:`ask`."""
        return self._batch_size

    def ask(self):
        """Creates solutions by adding Gaussian noise to elites in the archive.

        If the archive is empty and ``self._initial_solutions`` is set, we
        return ``self._initial_solutions``. If ``self._initial_solutions`` is
        not set, we draw from Gaussian distribution centered at ``self.x0``
        with standard deviation ``self.sigma``. Otherwise, each solution is
        drawn from a distribution centered at a randomly chosen elite with
        standard deviation ``self.sigma``.

        Returns:
            If the archive is not empty, ``(batch_size, solution_dim)`` array
            -- contains ``batch_size`` new solutions to evaluate. If the
            archive is empty, we return ``self._initial_solutions``, which
            might not have ``batch_size`` solutions.
        """
        if self.archive.empty:
            if self._initial_solutions is not None:
                return np.clip(self._initial_solutions, self.lower_bounds,
                               self.upper_bounds)
            parents = np.expand_dims(self.x0, axis=0)
        else:
            parents = self.archive.sample_elites(
                self._batch_size).solution_batch

        noise = self._rng.normal(
            scale=self._sigma,
            size=(self._batch_size, self.solution_dim),
        ).astype(self.archive.dtype)

        return np.clip(parents + noise, self.lower_bounds, self.upper_bounds)



"""Miscellaneous internal utilities."""
import numpy as np


def check_finite(x, name):
    """Checks that x is finite (i.e. not infinity or NaN).

    `x` must be either a scalar or NumPy array.
    """
    if not np.all(np.isfinite(x)):
        if np.isscalar(x):
            raise ValueError(f"{name} must be finite (infinity "
                             "and NaN values are not supported).")
        raise ValueError(f"All elements of {name} must be finite (infinity "
                         "and NaN values are not supported).")


def check_batch_shape(array, array_name, dim, dim_name, extra_msg=""):
    """Checks that the array has shape (batch_size, dim).

    `batch_size` can be any value.

    `array` must be a numpy array, and `dim` must be an int.
    """
    if array.ndim != 2 or array.shape[1] != dim:
        raise ValueError(f"Expected {array_name} to be a 2D array with shape "
                         f"(batch_size, {dim}) (i.e. shape "
                         f"(batch_size, {dim_name})) but it had shape "
                         f"{array.shape}.{extra_msg}")


def check_1d_shape(array, array_name, dim, dim_name, extra_msg=""):
    """Checks that the array has shape (dim,).

    `array` must be a numpy array, and `dim` must be an int.
    """
    if array.ndim != 1 or array.shape[0] != dim:
        raise ValueError(
            f"Expected {array_name} to be a 1D array with shape "
            f"({dim},) (i.e. shape ({dim_name},)) but it had shape "
            f"{array.shape}.{extra_msg}")


def check_is_1d(array, array_name, extra_msg=""):
    """Checks that an array is 1D."""
    if array.ndim != 1:
        raise ValueError(f"Expected {array_name} to be a 1D array but it had "
                         f"shape {array.shape}.{extra_msg}")


def check_batch_shape_3d(array,
                         array_name,
                         dim1,
                         dim1_name,
                         dim2,
                         dim2_name,
                         extra_msg=""):
    """Checks that the array has shape (batch_size, dim1, dim2).

    `batch_size` can be any value.

    `array` must be a numpy array, `dim1` and `dim2` must be int.
    """
    if array.ndim != 3 or array.shape[1] != dim1 or array.shape[2] != dim2:
        raise ValueError(f"Expected {array_name} to be a 3D array with shape "
                         f"(batch_size, {dim1}, {dim2}) (i.e. shape "
                         f"(batch_size, {dim1_name}, {dim2_name})) but it had"
                         f"shape {array.shape}.{extra_msg}")


def check_solution_batch_dim(array,
                             array_name,
                             batch_size,
                             is_1d=False,
                             extra_msg=""):
    """Checks the batch dimension of an array with respect to solution_batch."""
    if array.shape[0] != batch_size:
        raise ValueError(f"{array_name} does not match the batch dimension of "
                         "solution_batch -- since solution_batch has shape "
                         f"({batch_size}, ..), {array_name} should have shape "
                         f"({batch_size},{'' if is_1d else ' ..'}), but it has "
                         f"shape {array.shape}.{extra_msg}")


_BATCH_WARNING = (" Note that starting in pyribs 0.5.0, add() and tell() take"
                  " in a batch of solutions unlike in pyribs 0.4.0, where add()"
                  " and tell() only took in a single solution.")


def validate_batch_args(archive,
                        *,
                        solution_batch,
                        objective_batch,
                        measures_batch,
                        status_batch=None,
                        value_batch=None,
                        jacobian_batch=None,
                        metadata_batch=None):
    """Performs checks for arguments to add() and tell()."""
    check_batch_shape(solution_batch, "solution_batch", archive.solution_dim,
                      "solution_dim", _BATCH_WARNING)
    batch_size = solution_batch.shape[0]

    check_is_1d(objective_batch, "objective_batch", _BATCH_WARNING)
    check_solution_batch_dim(objective_batch,
                             "objective_batch",
                             batch_size,
                             is_1d=True,
                             extra_msg=_BATCH_WARNING)
    check_finite(objective_batch, "objective_batch")

    check_batch_shape(measures_batch, "measures_batch", archive.measure_dim,
                      "measure_dim", _BATCH_WARNING)
    check_solution_batch_dim(measures_batch,
                             "measures_batch",
                             batch_size,
                             is_1d=False,
                             extra_msg=_BATCH_WARNING)
    check_finite(measures_batch, "measures_batch")

    if jacobian_batch is not None:
        check_batch_shape_3d(jacobian_batch, "jacobian_batch",
                             archive.measure_dim + 1, "measure_dim + 1",
                             archive.solution_dim, "solution_dim")
        check_finite(jacobian_batch, "jacobian_batch")

    if status_batch is not None:
        check_is_1d(status_batch, "status_batch", _BATCH_WARNING)
        check_solution_batch_dim(status_batch,
                                 "status_batch",
                                 batch_size,
                                 is_1d=True,
                                 extra_msg=_BATCH_WARNING)
        check_finite(status_batch, "status_batch")

    if value_batch is not None:
        check_is_1d(value_batch, "value_batch", _BATCH_WARNING)
        check_solution_batch_dim(value_batch,
                                 "value_batch",
                                 batch_size,
                                 is_1d=True,
                                 extra_msg=_BATCH_WARNING)

    if metadata_batch is not None:
        check_is_1d(metadata_batch, "metadata_batch", _BATCH_WARNING)
        check_solution_batch_dim(metadata_batch,
                                 "metadata_batch",
                                 batch_size,
                                 is_1d=True,
                                 extra_msg=_BATCH_WARNING)


def validate_single_args(archive, solution, objective, measures):
    """Performs preprocessing and checks for arguments to add_single()."""
    check_1d_shape(solution, "solution", archive.solution_dim, "solution_dim")

    check_finite(objective, "objective")

    check_1d_shape(measures, "measures", archive.measure_dim, "measure_dim")
    check_finite(measures, "measures")

def readonly(arr):
    """Sets an array to be readonly."""
    arr.flags.writeable = False
    return arr



"""Provides GradientOptBase."""
from abc import ABC, abstractmethod


class GradientOptBase(ABC):
    """Base class for gradient-based optimizers.

    .. note::
        These optimizers are designed for gradient ascent rather than gradient
        descent.

    These optimizers maintain a current solution point :math:`\\theta`. The
    solution point is obtained with the :attr:`theta` property, and it is
    updated by passing a gradient to :meth:`step`. Finally, the point can be
    reset to a new value with :meth:`reset`.

    Your constructor may take in additional arguments beyond ``theta0`` and
    ``lr``, but expect that these two arguments will always be passed in.

    Args:
        theta0 (array-like): Initial solution. 1D array.
        lr (float): Learning rate for the update.
    """

    def __init__(self, theta0, lr):
        pass

    @property
    @abstractmethod
    def theta(self):
        """The current solution point."""

    @abstractmethod
    def reset(self, theta0):
        """Resets the solution point to a new value.

        Args:
            theta0 (array-like): The new solution point. 1D array.
        """

    @abstractmethod
    def step(self, gradient):
        """Ascends the solution based on the given gradient.

        Args:
            gradient (array-like): The (estimated) gradient of the current
                solution point. 1D array.
        """



"""Various optimizers which are employed across emitters.

There are gradient-based optimizers which inherit from :class:`GradientOptBase`
and evolution strategies which inherit from :class:`EvolutionStrategyBase`. When
specifying optimizers for an emitter, one can pass in the optimizer class
itself, or the string name of the optimizer, or an abbreviated name. The
supported abbreviations are as follows.

For gradient-based optimizers (inheriting from :class:`GradientOptBase`):

* ``adam``: :class:`AdamOpt`
* ``gradient_ascent``: :class:`GradientAscentOpt`

For evolution strategies (inheriting from :class:`EvolutionStrategyBase`):

* ``cma_es``: :class:`CMAEvolutionStrategy`
* ``sep_cma_es``: :class:`SeparableCMAEvolutionStrategy`
* ``lm_ma_es``: :class:`LMMAEvolutionStrategy`
* ``openai_es``: :class:`OpenAIEvolutionStrategy`

.. autosummary::
    :toctree:

    ribs.emitters.opt.CMAEvolutionStrategy
    ribs.emitters.opt.LMMAEvolutionStrategy
    ribs.emitters.opt.OpenAIEvolutionStrategy
    ribs.emitters.opt.SeparableCMAEvolutionStrategy
    ribs.emitters.opt.EvolutionStrategyBase
    ribs.emitters.opt.AdamOpt
    ribs.emitters.opt.GradientAscentOpt
    ribs.emitters.opt.GradientOptBase
"""
from ribs.emitters.opt._adam_opt import AdamOpt
from ribs.emitters.opt._cma_es import CMAEvolutionStrategy
from ribs.emitters.opt._evolution_strategy_base import EvolutionStrategyBase
from ribs.emitters.opt._gradient_ascent_opt import GradientAscentOpt
from ribs.emitters.opt._gradient_opt_base import GradientOptBase
from ribs.emitters.opt._lm_ma_es import LMMAEvolutionStrategy
from ribs.emitters.opt._openai_es import OpenAIEvolutionStrategy
from ribs.emitters.opt._sep_cma_es import SeparableCMAEvolutionStrategy

__all__ = [
    "CMAEvolutionStrategy",
    "LMMAEvolutionStrategy",
    "OpenAIEvolutionStrategy",
    "SeparableCMAEvolutionStrategy",
    "EvolutionStrategyBase",
    "AdamOpt",
    "GradientAscentOpt",
    "GradientOptBase",
]

_NAME_TO_GRAD_OPT_MAP = {
    "AdamOpt": AdamOpt,
    "GradientAscentOpt": GradientAscentOpt,
    "adam": AdamOpt,
    "gradient_ascent": GradientAscentOpt,
}


def _get_grad_opt(klass, **grad_opt_kwargs):
    """Returns a gradient optimizer class based on its name.

    Args:
        klass: Either a callable or a str for the gradient optimizer.
        grad_opt_kwargs (dict): Additional kwargs for the gradient optimizer.

    Returns:
        The corresponding gradient optimizer class instance.
    """
    if isinstance(klass, str):
        if klass in _NAME_TO_GRAD_OPT_MAP:
            klass = _NAME_TO_GRAD_OPT_MAP[klass]
        else:
            raise ValueError(f"`{klass}` is not the full or abbreviated "
                             "name of a valid gradient optimizer")
    if callable(klass):
        grad_opt = klass(**grad_opt_kwargs)
        if isinstance(grad_opt, GradientOptBase):
            return grad_opt
        raise ValueError(f"Callable `{klass}` did not return an instance "
                         "of GradientOptBase.")
    raise ValueError(f"`{klass}` is neither a callable nor a string")


_NAME_TO_ES_MAP = {
    "CMAEvolutionStrategy": CMAEvolutionStrategy,
    "SeparableCMAEvolutionStrategy": SeparableCMAEvolutionStrategy,
    "LMMAEvolutionStrategy": LMMAEvolutionStrategy,
    "OpenAIEvolutionStrategy": OpenAIEvolutionStrategy,
    "cma_es": CMAEvolutionStrategy,
    "sep_cma_es": SeparableCMAEvolutionStrategy,
    "lm_ma_es": LMMAEvolutionStrategy,
    "openai_es": OpenAIEvolutionStrategy,
}


def _get_es(klass, **es_kwargs):
    """Returns an evolution strategy (ES) class based on its name.

    Args:
        klass: Either a callable or a str for the ES.
        es_kwargs (dict): Additional keyword arguments for the ES.

    Returns:
        The corresponding evolution strategy class instance.
    """
    if isinstance(klass, str):
        if klass in _NAME_TO_ES_MAP:
            klass = _NAME_TO_ES_MAP[klass]
        else:
            raise ValueError(f"`{klass}` is not the full or abbreviated "
                             "name of a valid evolution strategy")
    if callable(klass):
        es = klass(**es_kwargs)
        if isinstance(es, EvolutionStrategyBase):
            return es
        raise ValueError(f"Callable `{klass}` did not return an instance "
                         "of EvolutionStrategyBase.")
    raise ValueError(f"`{klass}` is neither a callable nor a string")



"""Provides the AddStatus for all archives to use."""
from enum import IntEnum


class AddStatus(IntEnum):
    """A status returned by the :meth:`~ArchiveBase.add` method in an archive.

    This class is an :class:`~enum.IntEnum` with the following values:

    - ``NOT_ADDED``: The solution given to :meth:`~ArchiveBase.add` was not
      added to the archive.
    - ``IMPROVE_EXISTING``: The solution given to :meth:`~ArchiveBase.add`
      improved an elite already in the archive.
    - ``NEW``: The solution given to :meth:`~ArchiveBase.add` created a new
      elite in the archive.

    Example:

        Check the status of an add operation as follows (note that these
        examples use :meth:`~ArchiveBase.add_single` rather than
        :meth:`~ArchiveBase.add`)::

            from ribs.archives import AddStatus
            status, _ = archive.add_single(solution, objective, measures)
            if status == AddStatus.NEW:
                # Do something if the solution made a new elite in the archive.

        To check whether the solution was added to the archive, the status can
        act like a bool::

            from ribs.archives import AddStatus
            status, _ = archive.add_single(solution, objective, measures)
            if status:
                # Do something if the solution was added to the archive.

        Finally, there is an ordering on statuses::

            AddStatus.NEW > AddStatus.IMPROVE_EXISTING > AddStatus.NOT_ADDED
    """
    NOT_ADDED = 0
    IMPROVE_EXISTING = 1
    NEW = 2



"""Provides the Bandit Scheduler."""
import warnings

import numpy as np

from ribs.schedulers._scheduler import Scheduler


class BanditScheduler:
    """Schedules emitters with a bandit algorithm.

    This implementation is based on `Cully 2021
    <https://arxiv.org/abs/2007.05352>`_.

    .. note::
        This class follows the similar ask-tell framework as
        :class:`Scheduler`, and enforces similar constraints in the arguments
        and methods. Please refer to the documentation of :class:`Scheduler`
        for more details.

    .. note::
        The main difference between :class:`BanditScheduler` and
        :class:`Scheduler` is that, unlike :class:`Scheduler`, DQD emitters are
        not supported by :class:`BanditScheduler`.

    To initialize this class, first create an archive and a list of emitters
    for the QD algorithm. The BanditScheduler will schedule the emitters using
    the Upper Confidence Bound - 1 algorithm (UCB1). Everytime :meth:`ask` is
    called, the emitters are sorted based on the potential reward function from
    UCB1. Then, the top `num_active` emitters are used for ask-tell.

    Args:
        archive (ribs.archives.ArchiveBase): An archive object, e.g. one
            selected from :mod:`ribs.archives`.
        emitter_pool (list of ribs.archives.EmitterBase): A pool of emitters to
            select from, e.g. :class:`ribs.emitters.GaussianEmitter`. On the
            first iteration, the first `num_active` emitters from the
            emitter_pool will be activated.
        num_active (int): The number of active emitters at a time. Active
            emitters are used when calling ask-tell.
        zeta (float): Hyperparamter of UCB1 that balances the trade-off between
            the accuracy and the uncertainty of the emitters. Increasing this
            parameter will emphasize the uncertainty of the emitters. Refer to
            the original paper for more information.
        reselect (str): Indicates how emitters are reselected from the pool.
            The default is "terminated", where only terminated/restarted
            emitters are deactivated and reselected (but they might be selected
            again). Alternatively, use "all" to reselect all active emitters
            every iteration.
        add_mode (str): Indicates how solutions should be added to the archive.
            The default is "batch", which adds all solutions with one call to
            :meth:`~ribs.archives.ArchiveBase.add`. Alternatively, use "single"
            to add the solutions one at a time with
            :meth:`~ribs.archives.ArchiveBase.add_single`. "single" mode is
            included for legacy reasons, as it was the only mode of operation
            in pyribs 0.4.0 and before. We highly recommend using "batch" mode
            since it is significantly faster.
        result_archive (ribs.archives.ArchiveBase): In some algorithms, such as
            CMA-MAE, the archive does not store all the best-performing
            solutions. The `result_archive` is a secondary archive where we can
            store all the best-performing solutions.
    Raises:
        ValueError: Number of active emitter is less than one.
        ValueError: Less emitters in the pool than the number of active
            emitters.
        ValueError: The emitters passed in do not have the same solution
            dimensions.
        ValueError: The same emitter instance was passed in multiple times.
            Each emitter should be a unique instance (see the warning above).
        ValueError: Invalid value for `add_mode`.
    """

    def __init__(self,
                 archive,
                 emitter_pool,
                 num_active,
                 *,
                 reselect="terminated",
                 zeta=0.05,
                 result_archive=None,
                 add_mode="batch"):
        if num_active < 1:
            raise ValueError("num_active cannot be less than 1.")

        if len(emitter_pool) < num_active:
            raise ValueError(f"The emitter pool must contain at least"
                             f"num_active emitters, but only"
                             f"{len(emitter_pool)} are given.")

        emitter_ids = set(id(e) for e in emitter_pool)
        if len(emitter_ids) != len(emitter_pool):
            raise ValueError(
                "Not all emitters passed in were unique (i.e. some emitters "
                "had the same id). If emitters were created with something "
                "like [EmitterClass(...)] * n, instead use "
                "[EmitterClass(...) for _ in range(n)] so that all emitters "
                "are unique instances.")

        self._solution_dim = emitter_pool[0].solution_dim

        for idx, emitter in enumerate(emitter_pool[1:]):
            if emitter.solution_dim != self._solution_dim:
                raise ValueError(
                    "All emitters must have the same solution dim, but "
                    f"Emitter {idx} has dimension {emitter.solution_dim}, "
                    f"while Emitter 0 has dimension {self._solution_dim}")

        if reselect not in ["terminated", "all"]:
            raise ValueError("add_mode must either be 'terminated' or 'all',"
                             f"but it was '{reselect}'")

        if add_mode not in ["single", "batch"]:
            raise ValueError("add_mode must either be 'batch' or 'single', but "
                             f"it was '{add_mode}'")

        if archive is result_archive:
            raise ValueError("`archive` has same id as `result_archive` -- "
                             "Note that `Scheduler.result_archive` already "
                             "defaults to be the same as `archive` if you pass "
                             "`result_archive=None`")

        self._archive = archive
        self._emitter_pool = np.array(emitter_pool)
        self._num_active = num_active
        self._add_mode = add_mode
        self._result_archive = result_archive
        self._reselect = reselect

        # Boolean mask of the active emitters. Initializes to the first
        # num_active emitters in the emitter pool.
        self._active_arr = np.zeros_like(self._emitter_pool, dtype=bool)

        # Used by UCB1 to select emitters.
        self._success = np.zeros_like(self._emitter_pool, dtype=float)
        self._selection = np.zeros_like(self._emitter_pool, dtype=float)
        self._restarts = np.zeros_like(self._emitter_pool, dtype=int)
        self._zeta = zeta

        # Keeps track of whether the scheduler should be receiving a call to
        # ask() or tell().
        self._last_called = None
        # The last set of solutions returned by ask().
        self._solution_batch = []
        # The number of solutions created by each emitter.
        self._num_emitted = np.array([None for _ in self._active_arr])

    @property
    def archive(self):
        """ribs.archives.ArchiveBase: Archive for storing solutions found in
        this scheduler."""
        return self._archive

    @property
    def emitters(self):
        """list of ribs.archives.EmitterBase: Emitters for generating solutions
        in this scheduler."""
        return self._active_arr

    @property
    def result_archive(self):
        """ribs.archives.ArchiveBase: Another archive for storing solutions
        found in this optimizer.
        If `result_archive` was not passed to the constructor, this property is
        the same as :attr:`archive`.
        """
        return (self._archive
                if self._result_archive is None else self._result_archive)

    def ask_dqd(self):
        """Generates a batch of solutions by calling ask_dqd() on all DQD
        emitters.

        This method is not supported for this scheduler and throws an error if
        called.

        Raises:
            NotImplementedError: This method is not supported by this
                scheduler.
        """
        raise NotImplementedError("ask_dqd() is not supported by"
                                  "BanditScheduler.")

    def ask(self):
        """Generates a batch of solutions by calling ask() on all active
        emitters.

        The emitters used by ask are determined by the UCB1 algorithm. Briefly,
        emitters that have never been selected before are prioritized, then
        emitters are sorted in descending order based on the accurary of their
        past prediction.

        .. note:: The order of the solutions returned from this method is
            important, so do not rearrange them.

        Returns:
            (batch_size, dim) array: An array of n solutions to evaluate. Each
            row contains a single solution.
        Raises:
            RuntimeError: This method was called without first calling
                :meth:`tell`.
        """
        if self._last_called == "ask":
            raise RuntimeError("ask cannot be called immediately after " +
                               self._last_called)
        self._last_called = "ask"

        if self._reselect == "terminated":
            # Reselect terminated emitters. Emitters are terminated if their
            # restarts attribute have incremented.
            emitter_restarts = np.array([
                emitter.restarts if hasattr(emitter, "restarts") else -1
                for emitter in self._emitter_pool
            ])
            reselect = emitter_restarts > self._restarts

            # If the emitter does not have "restarts" attribute, assume it
            # restarts every iteration.
            reselect[emitter_restarts < 0] = True

            self._restarts = emitter_restarts
        else:
            # Reselect all emitters.
            reselect = self._active_arr.copy()

        # If no emitters are active, activate the first num_active.
        if not self._active_arr.any():
            reselect[:] = False
            self._active_arr[:self._num_active] = True

        # Deactivate emitters to be reselected.
        self._active_arr[reselect] = False

        # Select emitters based on the UCB1 formula.
        # The ranking of emitters also follows these rules:
        # - Emitters that have never been selected are prioritized.
        # - If reselect is "terminated", then only active emitters that have
        #   terminated/restarted will be reselected. Otherwise, if reselect is
        #   "all", then all emitters are reselected.
        if reselect.any():
            ucb1 = np.full_like(self._emitter_pool, np.inf)
            update_ucb = (self._selection != 0)
            if update_ucb.any():
                ucb1[update_ucb] = (
                    self._success[update_ucb] / self._selection[update_ucb] +
                    self._zeta * np.sqrt(
                        np.log(self._success.sum()) /
                        self._selection[update_ucb]))
            # Activate top emitters based on UCB1.
            activate = np.argsort(ucb1)[-reselect.sum():]
            self._active_arr[activate] = True

        self._solution_batch = []

        for i in np.where(self._active_arr)[0]:
            emitter = self._emitter_pool[i]
            emitter_sols = emitter.ask()
            self._solution_batch.append(emitter_sols)
            self._num_emitted[i] = len(emitter_sols)

        # In case the emitters didn't return any solutions.
        self._solution_batch = np.concatenate(
            self._solution_batch, axis=0) if self._solution_batch else np.empty(
                (0, self._solution_dim))
        return self._solution_batch

    def _check_length(self, name, array):
        """Raises a ValueError if array does not have the same length as the
        solutions."""
        if len(array) != len(self._solution_batch):
            raise ValueError(
                f"{name} should have length {len(self._solution_batch)} "
                "(this is the number of solutions output by ask()) but "
                f"has length {len(array)}")

    def tell_dqd(self,
                 objective_batch,
                 measures_batch,
                 jacobian_batch,
                 metadata_batch=None):
        """Returns info for solutions from :meth:`ask_dqd`.

        This method is not supported for this scheduler and throws an error if
        called.

        Raises:
            NotImplementedError: This method is not supported by this
                scheduler.
        """
        raise NotImplementedError("tell_dqd() is not supported by"
                                  "BanditScheduler.")

    def tell(self, objective_batch, measures_batch, metadata_batch=None):
        """Returns info for solutions from :meth:`ask`.

        The emitters are the same with those used in the last call to
        :meth:`ask`.

        .. note:: The objective batch, measures batch, and metadata batch must
            be in the same order as the solutions created by :meth:`ask`;
            i.e.  ``objective_batch[i]``, ``measures_batch[i]``, and
            ``metadata_batch[i]`` should be the objective, measures, and
            metadata for ``solution_batch[i]``.

        Args:
            objective_batch ((batch_size,) array): Each entry of this array
                contains the objective function evaluation of a solution.
            measures_batch ((batch_size, measures_dm) array): Each row of
                this array contains a solution's coordinates in measure space.
            metadata_batch ((batch_size,) array): Each entry of this array
                contains an object holding metadata for a solution.
        Raises:
            RuntimeError: This method is called without first calling
                :meth:`ask`.
            ValueError: ``objective_batch``, ``measures_batch``, or
                ``metadata_batch`` has the wrong shape.
        """
        if self._last_called != "ask":
            raise RuntimeError("tell() was called without calling ask().")
        self._last_called = "tell"

        objective_batch = np.asarray(objective_batch)
        measures_batch = np.asarray(measures_batch)
        metadata_batch = (np.empty(len(self._solution_batch), dtype=object) if
                          metadata_batch is None else np.asarray(metadata_batch,
                                                                 dtype=object))

        self._check_length("objective_batch", objective_batch)
        self._check_length("measures_batch", measures_batch)
        self._check_length("metadata_batch", metadata_batch)

        archive_empty_before = self.archive.empty
        if self._result_archive is not None:
            # Check self._result_archive here since self.result_archive is a
            # property that always provides a proper archive.
            result_archive_empty_before = self.result_archive.empty

        # Add solutions to the archive.
        if self._add_mode == "batch":
            status_batch, value_batch = self.archive.add(
                self._solution_batch,
                objective_batch,
                measures_batch,
                metadata_batch,
            )

            # Add solutions to result_archive.
            if self._result_archive is not None:
                self._result_archive.add(self._solution_batch, objective_batch,
                                         measures_batch, metadata_batch)
        elif self._add_mode == "single":
            status_batch = []
            value_batch = []
            for solution, objective, measure, metadata in zip(
                    self._solution_batch, objective_batch, measures_batch,
                    metadata_batch):
                status, value = self.archive.add_single(solution, objective,
                                                        measure, metadata)
                status_batch.append(status)
                value_batch.append(value)

                # Add solutions to result_archive.
                if self._result_archive is not None:
                    self._result_archive.add_single(solution, objective,
                                                    measure, metadata)
            status_batch = np.asarray(status_batch)
            value_batch = np.asarray(value_batch)

        # Warn the user if nothing was inserted into the archives.
        if archive_empty_before and self.archive.empty:
            warnings.warn(Scheduler.EMPTY_WARNING.format(name="archive"))
        if self._result_archive is not None:
            if result_archive_empty_before and self.result_archive.empty:
                warnings.warn(
                    Scheduler.EMPTY_WARNING.format(name="result_archive"))

        # Keep track of pos because emitters may have different batch sizes.
        pos = 0
        for i in np.where(self._active_arr)[0]:
            emitter = self._emitter_pool[i]
            n = self._num_emitted[i]

            end = pos + n
            self._selection[i] += n
            self._success[i] += np.count_nonzero(status_batch[pos:end])
            emitter.tell(self._solution_batch[pos:end],
                         objective_batch[pos:end], measures_batch[pos:end],
                         status_batch[pos:end], value_batch[pos:end],
                         metadata_batch[pos:end])
            pos = end
